# Copyright (C) 2015-2021 Regents of the University of California
# Copyright (C) 2015 Curoverse, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections.abc import Generator
import json
import logging
import os
import re
import shutil
import stat
import subprocess
import sys
import uuid
import zipfile
from functools import partial
from io import StringIO
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Optional, cast
from unittest.mock import Mock, call
from urllib.request import urlretrieve

if TYPE_CHECKING:
    from cwltool.utils import CWLObjectType

import pytest

pkg_root = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))  # noqa
sys.path.insert(0, pkg_root)  # noqa

from schema_salad.exceptions import ValidationException

from toil.cwl.utils import (
    DirectoryStructure,
    download_structure,
    visit_cwl_class_and_reduce,
    visit_top_cwl_class,
)
from toil.fileStores import FileID
from toil.fileStores.abstractFileStore import AbstractFileStore
from toil.lib.threading import cpu_count
from toil.test import (
    get_data,
)
from toil.test import (
    pslow as slow,
    pneeds_docker as needs_docker,
    pneeds_cwl as needs_cwl,
    pneeds_aws_s3 as needs_aws_s3,
    pneeds_docker_cuda as needs_docker_cuda,
    pneeds_gridengine as needs_gridengine,
    pneeds_kubernetes as needs_kubernetes,
    pneeds_local_cuda as needs_local_cuda,
    pneeds_lsf as needs_lsf,
    pneeds_mesos as needs_mesos,
    pneeds_online as needs_online,
    pneeds_slurm as needs_slurm,
    pneeds_torque as needs_torque,
    pneeds_wes_server as needs_wes_server,
)

log = logging.getLogger(__name__)
CONFORMANCE_TEST_TIMEOUT = 10000


def run_conformance_tests(
    workDir: str,
    yml: str,
    runner: Optional[str] = None,
    caching: bool = False,
    batchSystem: Optional[str] = None,
    selected_tests: Optional[str] = None,
    selected_tags: Optional[str] = None,
    skipped_tests: Optional[str] = None,
    extra_args: Optional[list[str]] = None,
    must_support_all_features: bool = False,
    junit_file: Optional[str] = None,
) -> None:
    """
    Run the CWL conformance tests.

    :param workDir: Directory to run tests in.

    :param yml: CWL test list YML to run tests from.

    :param runner: If set, use this cwl runner instead of the default toil-cwl-runner.

    :param caching: If True, use Toil file store caching.

    :param batchSystem: If set, use this batch system instead of the default single_machine.

    :param selected_tests: If set, use this description of test numbers to run (comma-separated numbers or ranges)

    :param selected_tags: As an alternative to selected_tests, run tests with the given tags.

    :param skipped_tests: Comma-separated string labels of tests to skip.

    :param extra_args: Provide these extra arguments to runner for each test.

    :param must_support_all_features: If set, fail if some CWL optional features are unsupported.

    :param junit_file: JUnit XML file to write test info to.
    """
    try:
        if runner is None:
            runner = "toil-cwl-runner"
        cmd = [
            "cwltest",
            f"--tool={runner}",
            f"--test={yml}",
            "--timeout=2400",
            f"--basedir={workDir}",
        ]
        if selected_tests:
            cmd.append(f"-n={selected_tests}")
        if selected_tags:
            cmd.append(f"--tags={selected_tags}")
        if skipped_tests:
            cmd.append(f"-S{skipped_tests}")
        if junit_file:
            # Capture output for JUnit
            cmd.append("--junit-verbose")
            cmd.append(f"--junit-xml={junit_file}")
        else:
            # Otherwise dump all output to our output stream
            cmd.append("--verbose")

        args_passed_directly_to_runner = [
            "--clean=always",
            "--logDebug",
            "--statusWait=10",
            "--retryCount=2",
            "--relax-path-checks",
            # Defaults to 20s but we can't start hundreds of nodejs processes that fast on our CI potatoes
            "--eval-timeout=600",
            f"--caching={caching}",
        ]

        if extra_args:
            args_passed_directly_to_runner += extra_args

        if "SINGULARITY_DOCKER_HUB_MIRROR" in os.environ:
            args_passed_directly_to_runner.append(
                "--setEnv=SINGULARITY_DOCKER_HUB_MIRROR"
            )

        if batchSystem is None or batchSystem == "single_machine":
            # Make sure we can run on small machines
            args_passed_directly_to_runner.append("--scale=0.1")

        job_store_override = None

        if batchSystem == "kubernetes":
            # Run tests in parallel on Kubernetes.
            # We can throw a bunch at it at once and let Kubernetes schedule.
            # But we still want a local core for each.
            parallel_tests = max(min(cpu_count(), 8), 1)
        else:
            # Run tests in parallel on the local machine. Don't run too many
            # tests at once; we want at least a couple cores for each.
            # But we need to have at least a few going in parallel or we risk hitting our timeout.
            parallel_tests = max(int(cpu_count() / 2), 4)
        cmd.append(f"-j{parallel_tests}")

        if batchSystem:
            args_passed_directly_to_runner.append(f"--batchSystem={batchSystem}")
        cmd.extend(["--"] + args_passed_directly_to_runner)

        log.info("Running: '%s'", "' '".join(cmd))
        output_lines: list[str] = []
        try:
            child = subprocess.Popen(
                cmd, cwd=workDir, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
            )

            if child.stdout is not None:
                for line_bytes in child.stdout:
                    # Pass through all the logs
                    line_text = line_bytes.decode("utf-8", errors="replace").rstrip()
                    output_lines.append(line_text)
                    log.info(line_text)

            # Once it's done writing, amke sure it succeeded.
            child.wait()
            log.info("CWL tests finished with exit code %s", child.returncode)
            if child.returncode != 0:
                # Act like check_output and raise an error.
                raise subprocess.CalledProcessError(child.returncode, " ".join(cmd))
        finally:
            if job_store_override:
                # Clean up the job store we used for all the tests, if it is still there.
                subprocess.run(["toil", "clean", job_store_override])

    except subprocess.CalledProcessError as e:
        log.info("CWL test runner return code was unsuccessful")
        only_unsupported = False
        # check output -- if we failed but only have unsupported features, we're okay
        p = re.compile(
            r"(?P<failures>\d+) failures, (?P<unsupported>\d+) unsupported features"
        )

        for line_text in output_lines:
            m = p.search(line_text)
            if m:
                if int(m.group("failures")) == 0 and int(m.group("unsupported")) > 0:
                    only_unsupported = True
                    break
        if (not only_unsupported) or must_support_all_features:
            log.error(
                "CWL tests gave unacceptable output:\n%s", "\n".join(output_lines)
            )
            raise e
        log.info("Unsuccessful return code is OK")


TesterFuncType = Callable[[Path, Path, "CWLObjectType", Path], None]


@needs_cwl
@pytest.mark.cwl
class TestCWLWorkflow:
    """
    CWL tests included in Toil that don't involve the whole CWL conformance
    test suite. Tests Toil-specific functions like URL types supported for
    inputs.
    """

    def test_cwl_cmdline_input(self) -> None:
        """
        Test that running a CWL workflow with inputs specified on the command line passes.
        """
        from toil.cwl import cwltoil

        with get_data("test/cwl/conditional_wf.cwl") as cwlfile:
            args = [str(cwlfile), "--message", "str", "--sleep", "2"]
            st = StringIO()
            # If the workflow runs, it must have had options
            cwltoil.main(args, stdout=st)

    def _tester(
        self,
        cwlfile: Path,
        jobfile: Path,
        expect: "CWLObjectType",
        outdir: Path,
        out_name: str = "output",
        main_args: Optional[list[str]] = None,
    ) -> None:
        from toil.cwl import cwltoil

        st = StringIO()
        real_main_args = main_args or []
        real_main_args.extend(
            [
                "--logDebug",
                "--outdir",
                str(outdir),
                str(cwlfile),
                str(jobfile),
            ]
        )
        cwltoil.main(real_main_args, stdout=st)
        out = json.loads(st.getvalue())
        out.get(out_name, {}).pop("http://commonwl.org/cwltool#generation", None)
        out.get(out_name, {}).pop("nameext", None)
        out.get(out_name, {}).pop("nameroot", None)
        assert out == expect

        for k, v in expect.items():
            if (
                isinstance(v, dict)
                and "class" in v
                and v["class"] == "File"
                and "path" in v
            ):
                # This is a top-level output file.
                # None of our output files should be executable.
                assert os.path.exists(v["path"]) is True
                assert (os.stat(v["path"]).st_mode & stat.S_IXUSR) == 0

    def _debug_worker_tester(
        self, cwlfile: Path, jobfile: Path, expect: "CWLObjectType", outdir: Path
    ) -> None:
        from toil.cwl import cwltoil

        st = StringIO()
        cwltoil.main(
            [
                "--debugWorker",
                "--outdir",
                str(outdir),
                str(cwlfile),
                str(jobfile),
            ],
            stdout=st,
        )
        out = json.loads(st.getvalue())
        out["output"].pop("http://commonwl.org/cwltool#generation", None)
        out["output"].pop("nameext", None)
        out["output"].pop("nameroot", None)
        assert out == expect

    def revsort(
        self, cwl_filename: str, tester_fn: TesterFuncType, out_dir: Path
    ) -> None:
        with get_data(f"test/cwl/{cwl_filename}") as cwl_file:
            with get_data("test/cwl/revsort-job.json") as job_file:
                tester_fn(
                    cwl_file, job_file, self._expected_revsort_output(out_dir), out_dir
                )

    def revsort_no_checksum(
        self, cwl_filename: str, tester_fn: TesterFuncType, out_dir: Path
    ) -> None:
        with get_data(f"test/cwl/{cwl_filename}") as cwl_file:
            with get_data("test/cwl/revsort-job.json") as job_file:
                tester_fn(
                    cwl_file,
                    job_file,
                    self._expected_revsort_nochecksum_output(out_dir),
                    out_dir,
                )

    def download(self, inputs: str, tester_fn: TesterFuncType, out_dir: Path) -> None:
        with get_data(f"test/cwl/{inputs}") as input_location:
            with get_data("test/cwl/download.cwl") as cwl_file:
                tester_fn(
                    cwl_file,
                    input_location,
                    self._expected_download_output(out_dir),
                    out_dir,
                )

    def load_contents(
        self, inputs: str, tester_fn: TesterFuncType, out_dir: Path
    ) -> None:
        with get_data(f"test/cwl/{inputs}") as input_location:
            with get_data("test/cwl/load_contents.cwl") as cwl_file:
                tester_fn(
                    cwl_file,
                    input_location,
                    self._expected_load_contents_output(out_dir),
                    out_dir,
                )

    def download_directory(
        self, inputs: str, tester_fn: TesterFuncType, out_dir: Path
    ) -> None:
        with get_data(f"test/cwl/{inputs}") as input_location:
            with get_data("test/cwl/download_directory.cwl") as cwl_file:
                tester_fn(
                    cwl_file,
                    input_location,
                    self._expected_download_output(out_dir),
                    out_dir,
                )

    def download_subdirectory(
        self, inputs: str, tester_fn: TesterFuncType, out_dir: Path
    ) -> None:
        with get_data(f"test/cwl/{inputs}") as input_location:
            with get_data("test/cwl/download_subdirectory.cwl") as cwl_file:
                tester_fn(
                    cwl_file,
                    input_location,
                    self._expected_download_output(out_dir),
                    out_dir,
                )

    def test_mpi(self, tmp_path: Path) -> None:
        from toil.cwl import cwltoil

        stdout = StringIO()
        with get_data("test/cwl/mock_mpi/fake_mpi.yml") as mpi_config_file:
            with get_data("test/cwl/mpi_simple.cwl") as cwl_file:
                with get_data("test/cwl/mock_mpi/fake_mpi_run.py") as fake_mpi_run:
                    main_args = [
                        "--logDebug",
                        "--outdir",
                        str(tmp_path),
                        "--enable-dev",
                        "--enable-ext",
                        "--mpi-config-file",
                        str(mpi_config_file),
                        str(cwl_file),
                    ]
                    path = os.environ["PATH"]
                    os.environ["PATH"] = f"{path}:{fake_mpi_run.parent}"
                    cwltoil.main(main_args, stdout=stdout)
                    os.environ["PATH"] = path
                    stdout_text = stdout.getvalue()
                    assert "pids" in stdout_text
                    out = json.loads(stdout_text)
                    with open(
                        out.get("pids", {}).get("location")[len("file://") :]
                    ) as f:
                        two_pids = [int(i) for i in f.read().split()]
                    assert len(two_pids) == 2
                    assert isinstance(two_pids[0], int)
                    assert isinstance(two_pids[1], int)

    @needs_aws_s3
    @pytest.mark.aws_s3
    @pytest.mark.online
    def test_s3_as_secondary_file(self, tmp_path: Path) -> None:
        from toil.cwl import cwltoil

        stdout = StringIO()
        with get_data("test/cwl/s3_secondary_file.cwl") as cwl_file:
            with get_data("test/cwl/s3_secondary_file.json") as inputs_file:
                main_args = ["--outdir", str(tmp_path), str(cwl_file), str(inputs_file)]
                cwltoil.main(main_args, stdout=stdout)
                out = json.loads(stdout.getvalue())
                assert (
                    out["output"]["checksum"]
                    == "sha1$d14dd02e354918b4776b941d154c18ebc15b9b38"
                )

                assert out["output"]["size"] == 24
                with open(out["output"]["location"][len("file://") :]) as f:
                    assert f.read().strip() == "When is s4 coming out?"

    @needs_docker
    @pytest.mark.docker
    @pytest.mark.online
    def test_run_revsort(self, tmp_path: Path) -> None:
        self.revsort("revsort.cwl", self._tester, tmp_path)

    @needs_docker
    @pytest.mark.docker
    @pytest.mark.online
    def test_run_revsort_nochecksum(self, tmp_path: Path) -> None:
        self.revsort_no_checksum(
            "revsort.cwl",
            partial(self._tester, main_args=["--no-compute-checksum"]),
            tmp_path,
        )

    def test_run_revsort_no_container(self, tmp_path: Path) -> None:
        self.revsort(
            "revsort.cwl", partial(self._tester, main_args=["--no-container"]), tmp_path
        )

    @needs_docker
    @pytest.mark.docker
    @pytest.mark.online
    def test_run_revsort2(self, tmp_path: Path) -> None:
        self.revsort("revsort2.cwl", self._tester, tmp_path)

    @needs_docker
    @pytest.mark.docker
    @pytest.mark.online
    def test_run_revsort_debug_worker(self, tmp_path: Path) -> None:
        self.revsort("revsort.cwl", self._debug_worker_tester, tmp_path)

    @needs_docker
    @pytest.mark.docker
    @pytest.mark.online
    def test_run_colon_output(self, tmp_path: Path) -> None:
        with get_data("test/cwl/colon_test_output.cwl") as cwl_file:
            with get_data("test/cwl/colon_test_output_job.yaml") as inputs_file:
                self._tester(
                    cwl_file,
                    inputs_file,
                    self._expected_colon_output(tmp_path),
                    tmp_path,
                    out_name="result",
                )

    @pytest.mark.integrative
    @needs_docker
    @pytest.mark.docker
    @pytest.mark.online
    def test_run_dockstore_trs(self, tmp_path: Path) -> None:
        from toil.cwl import cwltoil

        stdout = StringIO()
        main_args = [
            "--outdir",
            str(tmp_path),
            "#workflow/github.com/dockstore-testing/md5sum-checker:master",
            "https://raw.githubusercontent.com/dockstore-testing/md5sum-checker/refs/heads/master/md5sum/md5sum-input-cwl.json",
        ]
        cwltoil.main(main_args, stdout=stdout)
        out = json.loads(stdout.getvalue())
        with open(out.get("output_file", {}).get("location")[len("file://") :]) as f:
            computed_hash = f.read().strip()
        assert computed_hash == "00579a00e3e7fa0674428ac7049423e2"

    def test_glob_dir_bypass_file_store(self, tmp_path: Path) -> None:
        self.maxDiff = 1000
        with get_data("test/cwl/glob_dir.cwl") as cwl_file:
            with get_data("test/cwl/empty.json") as inputs_file:
                self._tester(
                    cwl_file,
                    inputs_file,
                    self._expected_glob_dir_output(tmp_path),
                    tmp_path,
                    main_args=["--bypass-file-store"],
                )

    def test_required_input_condition_protection(self, tmp_path: Path) -> None:
        # This doesn't run containerized
        with get_data("test/cwl/not_run_required_input.cwl") as cwl_file:
            with get_data("test/cwl/empty.json") as inputs_file:
                self._tester(cwl_file, inputs_file, {}, tmp_path)

    @needs_slurm
    @pytest.mark.slurm
    def test_slurm_node_memory(self, tmp_path: Path) -> None:
        pass

        # Run the workflow. This will either finish quickly and tell us the
        # memory we got, or take a long time because it requested a whole
        # node's worth of memory and no nodes are free right now. We need to
        # support both.

        # And if we run out of time we need to stop the workflow gracefully and
        # cancel the Slurm jobs.

        try:
            with get_data("test/cwl/measure_default_memory.cwl") as cwl_file:
                main_args = [
                    f"--jobStore={str(tmp_path / 'jobStoreDir')}",
                    # Avoid racing to toil kill before the jobstore is removed
                    "--clean=never",
                    "--batchSystem=slurm",
                    "--no-cwl-default-ram",
                    "--slurmDefaultAllMem=True",
                    "--outdir",
                    str(tmp_path / "outdir"),
                    str(cwl_file),
                ]
                log.debug("Start test workflow")
                child = subprocess.Popen(
                    ["toil-cwl-runner"] + main_args, stdout=subprocess.PIPE
                )
                output, _ = child.communicate(timeout=60)
        except subprocess.TimeoutExpired:
            # The job didn't finish quickly; presumably waiting for a full node.
            # Stop the workflow
            log.debug("Workflow might be waiting for a full node. Stop it.")
            subprocess.check_call(["toil", "kill", str(tmp_path / "jobStoreDir")])
            # Wait another little bit for it to clean up, making sure to collect output in case it is blocked on writing
            child.communicate(timeout=20)
            # Kill it off in case it is still running
            child.kill()
            # Reap it
            child.wait()
            # The test passes
        else:
            out = json.loads(output)
            log.debug("Workflow output: %s", out)
            memory_string = out["memory"]
            log.debug("Observed memory: %s", memory_string)
            # If there's no memory limit enforced, Slurm will return "unlimited".
            # Set result to something sensible.
            if memory_string.strip() == "unlimited":
                result = 4 * 1024 * 1024
            else:
                result = int(memory_string)
            # We should see more than the CWL default or the Toil default, assuming Slurm nodes of reasonable size (3 GiB).
            assert result > (3 * 1024 * 1024)

    @needs_aws_s3
    @pytest.mark.aws_s3
    @pytest.mark.online
    def test_download_s3(self, tmp_path: Path) -> None:
        self.download("download_s3.json", self._tester, tmp_path)

    def test_download_http(self, tmp_path: Path) -> None:
        self.download("download_http.json", self._tester, tmp_path)

    def test_download_https(self, tmp_path: Path) -> None:
        self.download("download_https.json", self._tester, tmp_path)

    def test_download_https_reference(self, tmp_path: Path) -> None:
        self.download(
            "download_https.json",
            partial(self._tester, main_args=["--reference-inputs"]),
            tmp_path,
        )

    def test_download_file(self, tmp_path: Path) -> None:
        self.download("download_file.json", self._tester, tmp_path)

    @needs_aws_s3
    @pytest.mark.aws_s3
    @pytest.mark.online
    def test_download_directory_s3(self, tmp_path: Path) -> None:
        self.download_directory("download_directory_s3.json", self._tester, tmp_path)

    @needs_aws_s3
    @pytest.mark.aws_s3
    @pytest.mark.online
    def test_download_directory_s3_reference(self, tmp_path: Path) -> None:
        self.download_directory(
            "download_directory_s3.json",
            partial(self._tester, main_args=["--reference-inputs"]),
            tmp_path,
        )

    def test_download_directory_file(self, tmp_path: Path) -> None:
        self.download_directory("download_directory_file.json", self._tester, tmp_path)

    @needs_aws_s3
    @pytest.mark.aws_s3
    @pytest.mark.online
    def test_download_subdirectory_s3(self, tmp_path: Path) -> None:
        self.download_subdirectory(
            "download_subdirectory_s3.json", self._tester, tmp_path
        )

    def test_download_subdirectory_file(self, tmp_path: Path) -> None:
        self.download_subdirectory(
            "download_subdirectory_file.json", self._tester, tmp_path
        )

    # We also want to make sure we can run a bare tool with loadContents on the inputs, which requires accessing the input data early in the leader.

    @needs_aws_s3
    @pytest.mark.aws_s3
    @pytest.mark.online
    def test_load_contents_s3(self, tmp_path: Path) -> None:
        self.load_contents("download_s3.json", self._tester, tmp_path)

    def test_load_contents_http(self, tmp_path: Path) -> None:
        self.load_contents("download_http.json", self._tester, tmp_path)

    def test_load_contents_https(self, tmp_path: Path) -> None:
        self.load_contents("download_https.json", self._tester, tmp_path)

    def test_load_contents_file(self, tmp_path: Path) -> None:
        self.load_contents("download_file.json", self._tester, tmp_path)

    @slow
    @pytest.mark.integrative
    @pytest.mark.slow
    @pytest.mark.skip("Fails too often due to remote service")
    def test_bioconda(self, tmp_path: Path) -> None:
        with get_data("test/cwl/seqtk_seq.cwl") as cwl_file:
            with get_data("test/cwl/seqtk_seq_job.json") as inputs_file:
                self._tester(
                    cwl_file,
                    inputs_file,
                    self._expected_seqtk_output(tmp_path),
                    tmp_path,
                    main_args=["--beta-conda-dependencies"],
                    out_name="output1",
                )

    @needs_docker
    @pytest.mark.docker
    @pytest.mark.online
    def test_default_args(self, tmp_path: Path) -> None:
        with get_data("test/cwl/seqtk_seq.cwl") as cwl_file:
            with get_data("test/cwl/seqtk_seq_job.json") as inputs_file:
                self._tester(
                    cwl_file,
                    inputs_file,
                    self._expected_seqtk_output(tmp_path),
                    tmp_path,
                    main_args=[
                        "--default-container",
                        "quay.io/biocontainers/seqtk:1.4--he4a0461_1",
                    ],
                    out_name="output1",
                )

    @needs_docker
    @pytest.mark.docker
    @pytest.mark.integrative
    @pytest.mark.online
    @pytest.mark.skip(reason="Fails too often due to remote service")
    def test_biocontainers(self, tmp_path: Path) -> None:
        with get_data("test/cwl/seqtk_seq.cwl") as cwl_file:
            with get_data("test/cwl/seqtk_seq_job.json") as inputs_file:
                self._tester(
                    cwl_file,
                    inputs_file,
                    self._expected_seqtk_output(tmp_path),
                    tmp_path,
                    main_args=["--beta-use-biocontainers"],
                    out_name="output1",
                )

    @needs_docker
    @needs_docker_cuda
    @needs_local_cuda
    @pytest.mark.docker
    @pytest.mark.online
    @pytest.mark.docker_cuda
    @pytest.mark.local_cuda
    def test_cuda(self, tmp_path: Path) -> None:
        with get_data("test/cwl/nvidia_smi.cwl") as cwl_file:
            with get_data("test/cwl/empty.json") as inputs_file:
                self._tester(
                    cwl_file,
                    inputs_file,
                    {},
                    tmp_path,
                    out_name="result",
                )

    @slow
    @pytest.mark.slow
    def test_restart(self, tmp_path: Path) -> None:
        """
        Enable restarts with toil-cwl-runner -- run failing test, re-run correct test.
        Only implemented for single machine.
        """
        log.info("Running CWL Test Restart.  Expecting failure, then success.")
        from toil.cwl import cwltoil

        outDir = tmp_path / "outDir"
        outDir.mkdir()
        jobStore = tmp_path / "jobStore"
        with get_data("test/cwl/revsort.cwl") as cwl_file:
            with get_data("test/cwl/revsort-job.json") as job_file:
                cmd = [
                    "--outdir",
                    str(outDir),
                    "--jobStore",
                    str(jobStore),
                    "--no-container",
                    str(cwl_file),
                    str(job_file),
                ]

                # create a fake rev bin that actually points to the "date" binary
                cal_path = [
                    d
                    for d in os.environ["PATH"].split(":")
                    if os.path.exists(os.path.join(d, "date"))
                ][-1]
                os.symlink(
                    os.path.realpath(os.path.join(cal_path, "date")), outDir / "rev"
                )

                def path_with_bogus_rev() -> str:
                    # append to the front of the PATH so that we check there first
                    return f"{str(outDir)}:" + os.environ["PATH"]

                orig_path = os.environ["PATH"]
                # Force a failure by trying to use an incorrect version of `rev` from the PATH
                os.environ["PATH"] = path_with_bogus_rev()
                try:
                    subprocess.check_output(
                        ["toil-cwl-runner"] + cmd,
                        env=os.environ.copy(),
                        stderr=subprocess.STDOUT,
                    )
                    pytest.fail("Expected problem job with incorrect PATH did not fail")
                except subprocess.CalledProcessError:
                    pass
                # Finish the job with a correct PATH
                os.environ["PATH"] = orig_path
                cmd.insert(0, "--restart")
                cwltoil.main(cmd)
                # Should fail because previous job completed successfully
                try:
                    subprocess.check_output(
                        ["toil-cwl-runner"] + cmd,
                        env=os.environ.copy(),
                        stderr=subprocess.STDOUT,
                    )
                    pytest.fail("Restart with missing directory did not fail")
                except subprocess.CalledProcessError:
                    pass

    def test_caching(self, tmp_path: Path) -> None:
        log.info("Running CWL caching test.")
        from toil.cwl import cwltoil

        outDir = tmp_path / "outDir"
        cacheDir = tmp_path / "cacheDir"
        log_path = outDir / "log"

        with get_data("test/cwl/revsort.cwl") as cwl_file:
            with get_data("test/cwl/revsort-job.json") as job_file:
                cmd = [
                    "--outdir",
                    str(outDir),
                    "--jobStore",
                    str(tmp_path / "jobStore"),
                    "--clean=always",
                    "--no-container",
                    "--cachedir",
                    str(cacheDir),
                    str(cwl_file),
                    str(job_file),
                ]
                st = StringIO()
                ret = cwltoil.main(cmd, stdout=st)
                assert ret == 0
                # cwltool hashes certain steps into directories, ensure it exists
                # since cwltool caches per task and revsort has 2 cwl tasks, there should be 2 directories and 2 status files
                assert sum(1 for _ in cacheDir.iterdir()) == 4

                # Rerun the workflow to ensure there is a cache hit and that we don't rerun the tools
                st = StringIO()
                cmd = [
                    "--writeLogsFromAllJobs=True",
                    "--writeLogs",
                    str(log_path),
                ] + cmd
                ret = cwltoil.main(cmd, stdout=st)
                assert ret == 0

                # Ensure all of the worker logs are using their cached outputs
                for file in log_path.iterdir():
                    assert "Using cached output" in file.read_text(encoding="utf-8")

    @needs_aws_s3
    @pytest.mark.aws_s3
    @pytest.mark.online
    def test_streamable(
        self, tmp_path: Path, extra_args: Optional[list[str]] = None
    ) -> None:
        """
        Test that a file with 'streamable'=True is a named pipe.
        This is a CWL1.2 feature.
        """
        st = StringIO()
        outDir = tmp_path / "outDir"
        with get_data("test/cwl/stream.cwl") as cwlfile:
            with get_data("test/cwl/stream.json") as jobfile:
                out_name = "output"
                jobstore = f"--jobStore=aws:us-west-1:toil-stream-{uuid.uuid4()}"
                from toil.cwl import cwltoil

                args = [
                    "--logDebug",
                    "--outdir",
                    str(outDir),
                    jobstore,
                    str(cwlfile),
                    str(jobfile),
                ]
                if extra_args:
                    args = extra_args + args
                log.info("Run CWL run: %s", " ".join(args))
                cwltoil.main(args, stdout=st)
        out = json.loads(st.getvalue())
        out[out_name].pop("http://commonwl.org/cwltool#generation", None)
        out[out_name].pop("nameext", None)
        out[out_name].pop("nameroot", None)
        assert out == self._expected_streaming_output(outDir)
        with open(out[out_name]["location"][len("file://") :]) as f:
            assert f.read().strip() == "When is s4 coming out?"

    @needs_aws_s3
    @pytest.mark.aws_s3
    @pytest.mark.online
    def test_streamable_reference(self, tmp_path: Path) -> None:
        """
        Test that a streamable file is a stream even when passed around by URI.
        """
        self.test_streamable(tmp_path=tmp_path, extra_args=["--reference-inputs"])

    def test_preemptible(self, tmp_path: Path) -> None:
        """
        Tests that the http://arvados.org/cwl#UsePreemptible extension is supported.
        """
        from toil.cwl import cwltoil

        st = StringIO()
        out_name = "output"
        with get_data("test/cwl/preemptible.cwl") as cwlfile:
            with get_data("test/cwl/empty.json") as jobfile:
                args = [
                    "--outdir",
                    str(tmp_path / "outDir"),
                    str(cwlfile),
                    str(jobfile),
                ]
                cwltoil.main(args, stdout=st)
        out = json.loads(st.getvalue())
        out[out_name].pop("http://commonwl.org/cwltool#generation", None)
        out[out_name].pop("nameext", None)
        out[out_name].pop("nameroot", None)
        with open(out[out_name]["location"][len("file://") :]) as f:
            assert f.read().strip() == "hello"

    def test_preemptible_expression(self, tmp_path: Path) -> None:
        """
        Tests that the http://arvados.org/cwl#UsePreemptible extension is validated.
        """
        from toil.cwl import cwltoil

        st = StringIO()
        with get_data("test/cwl/preemptible_expression.cwl") as cwlfile:
            with get_data("test/cwl/preemptible_expression.json") as jobfile:
                args = [
                    "--outdir",
                    str(tmp_path),
                    str(cwlfile),
                    str(jobfile),
                ]
                with pytest.raises(
                    ValidationException, match=re.escape("expressions are not allowed")
                ):
                    cwltoil.main(args, stdout=st)

    @staticmethod
    def _expected_seqtk_output(outDir: Path) -> "CWLObjectType":
        path = outDir / "out"
        return {
            "output1": {
                "location": path.as_uri(),
                "path": str(path),
                "checksum": "sha1$322e001e5a99f19abdce9f02ad0f02a17b5066c2",
                "basename": "out",
                "class": "File",
                "size": 150,
            }
        }

    @staticmethod
    def _expected_revsort_output(outDir: Path) -> "CWLObjectType":
        path = outDir / "output.txt"
        return {
            "output": {
                "location": path.as_uri(),
                "path": str(path),
                "basename": "output.txt",
                "size": 1111,
                "class": "File",
                "checksum": "sha1$b9214658cc453331b62c2282b772a5c063dbd284",
            }
        }

    @staticmethod
    def _expected_revsort_nochecksum_output(outDir: Path) -> "CWLObjectType":
        path = outDir / "output.txt"
        return {
            "output": {
                "location": path.as_uri(),
                "path": str(path),
                "basename": "output.txt",
                "size": 1111,
                "class": "File",
            }
        }

    @staticmethod
    def _expected_download_output(outDir: Path) -> "CWLObjectType":
        path = outDir / "output.txt"
        return {
            "output": {
                "location": path.as_uri(),
                "basename": "output.txt",
                "size": 0,
                "class": "File",
                "checksum": "sha1$da39a3ee5e6b4b0d3255bfef95601890afd80709",
                "path": str(path),
            }
        }

    @staticmethod
    def _expected_glob_dir_output(out_dir: Path) -> "CWLObjectType":
        dir_path = out_dir / "shouldmake"
        dir_loc = dir_path.as_uri()
        file_path = dir_path / "test.txt"
        file_loc = file_path.as_uri()
        return {
            "shouldmake": {
                "location": dir_loc,
                "path": str(dir_path),
                "basename": "shouldmake",
                "nameroot": "shouldmake",
                "nameext": "",
                "class": "Directory",
                "listing": [
                    {
                        "class": "File",
                        "location": file_loc,
                        "path": str(file_path),
                        "basename": "test.txt",
                        "checksum": "sha1$da39a3ee5e6b4b0d3255bfef95601890afd80709",
                        "size": 0,
                        "nameroot": "test",
                        "nameext": ".txt",
                    }
                ],
            }
        }

    @classmethod
    def _expected_load_contents_output(cls, out_dir: Path) -> "CWLObjectType":
        """
        Generate the putput we expect from load_contents.cwl, when sending
        output files to the given directory.
        """
        expected = cls._expected_download_output(out_dir)
        expected["length"] = 146
        return expected

    @staticmethod
    def _expected_colon_output(outDir: Path) -> "CWLObjectType":
        path = outDir / "A:Gln2Cys_result"
        loc = "file://" + os.path.join(
            outDir, "A%3AGln2Cys_result"
        )  # not using .as_uri to ensure the expected escaping
        return {
            "result": {
                "location": loc,
                "path": str(path),
                "basename": "A:Gln2Cys_result",
                "class": "Directory",
                "listing": [
                    {
                        "class": "File",
                        "location": f"{loc}/whale.txt",
                        "basename": "whale.txt",
                        "checksum": "sha1$327fc7aedf4f6b69a42a7c8b808dc5a7aff61376",
                        "size": 1111,
                        "nameroot": "whale",
                        "nameext": ".txt",
                        "path": f"{path}/whale.txt",
                    }
                ],
            }
        }

    def _expected_streaming_output(self, outDir: Path) -> "CWLObjectType":
        path = outDir / "output.txt"
        return {
            "output": {
                "location": path.as_uri(),
                "path": str(path),
                "basename": "output.txt",
                "size": 24,
                "class": "File",
                "checksum": "sha1$d14dd02e354918b4776b941d154c18ebc15b9b38",
            }
        }

    @needs_docker
    @pytest.mark.docker
    @pytest.mark.online
    def test_missing_import(self, tmp_path: Path) -> None:
        with get_data("test/cwl/revsort.cwl") as cwl_file:
            with get_data("test/cwl/revsort-job-missing.json") as inputs_file:
                cmd = [
                    "toil-cwl-runner",
                    f"--outdir={str(tmp_path)}",
                    "--clean=always",
                    str(cwl_file),
                    str(inputs_file),
                ]
                p = subprocess.run(cmd, capture_output=True, text=True)
                # Make sure that the missing file is mentioned in the log so the user knows
                assert p.returncode == 1, p.stderr
                assert "missing.txt" in p.stderr

    @needs_aws_s3
    @pytest.mark.aws_s3
    @pytest.mark.online
    def test_optional_secondary_files_exists(self, tmp_path: Path) -> None:
        from toil.cwl import cwltoil

        with get_data("test/cwl/optional-file.cwl") as cwlfile:
            with get_data("test/cwl/optional-file-exists.json") as jobfile:
                args = [str(cwlfile), str(jobfile), f"--outdir={str(tmp_path)}"]
                ret = cwltoil.main(args)
                assert ret == 0
                assert (tmp_path / "wdl_templates_old.zip").exists()

    @needs_aws_s3
    @pytest.mark.aws_s3
    @pytest.mark.online
    def test_optional_secondary_files_missing(self, tmp_path: Path) -> None:
        from toil.cwl import cwltoil

        with get_data("test/cwl/optional-file.cwl") as cwlfile:
            with get_data("test/cwl/optional-file-missing.json") as jobfile:
                args = [str(cwlfile), str(jobfile), f"--outdir={str(tmp_path)}"]
                ret = cwltoil.main(args)
                assert ret == 0
                assert not (tmp_path / "hello_old.zip").exists()


@pytest.fixture(scope="function")
def cwl_v1_0_spec(tmp_path: Path) -> Generator[Path]:
    # The latest cwl git commit hash from https://github.com/common-workflow-language/common-workflow-language.
    # Update it to get the latest tests.
    testhash = (
        "6a955874ade22080b8ef962b4e0d6e408112c1ef"  # Date:   Tue Dec 16 2020 8:43pm PST
    )
    url = (
        "https://github.com/common-workflow-language/common-workflow-language/archive/%s.zip"
        % testhash
    )
    urlretrieve(url, "spec.zip")
    with zipfile.ZipFile("spec.zip", "r") as z:
        z.extractall()
    shutil.move("common-workflow-language-%s" % testhash, str(tmp_path))
    os.remove("spec.zip")
    try:
        yield tmp_path / ("common-workflow-language-%s" % testhash)
    finally:
        pass  # no cleanup


@needs_cwl
@needs_online
@pytest.mark.cwl
@pytest.mark.online
class TestCWLv10:
    """
    Run the CWL 1.0 conformance tests in various environments.
    """

    @slow
    @needs_docker
    @pytest.mark.slow
    @pytest.mark.docker
    @pytest.mark.online
    @pytest.mark.timeout(CONFORMANCE_TEST_TIMEOUT)
    def test_run_conformance_with_caching(self, cwl_v1_0_spec: Path) -> None:
        self.test_run_conformance(cwl_v1_0_spec, caching=True)

    @slow
    @needs_docker
    @pytest.mark.slow
    @pytest.mark.docker
    @pytest.mark.online
    @pytest.mark.timeout(CONFORMANCE_TEST_TIMEOUT)
    def test_run_conformance(
        self,
        cwl_v1_0_spec: Path,
        batchSystem: Optional[str] = None,
        caching: bool = False,
        selected_tests: Optional[str] = None,
        skipped_tests: Optional[str] = None,
        extra_args: Optional[list[str]] = None,
    ) -> None:
        run_conformance_tests(
            workDir=str(cwl_v1_0_spec / "v1.0"),
            yml=str(cwl_v1_0_spec / "v1.0" / "conformance_test_v1.0.yaml"),
            caching=caching,
            batchSystem=batchSystem,
            selected_tests=selected_tests,
            skipped_tests=skipped_tests,
            extra_args=extra_args,
        )

    @slow
    @needs_lsf
    @pytest.mark.slow
    @pytest.mark.lsf
    @pytest.mark.skip("Not run")
    def test_lsf_cwl_conformance(
        self, cwl_v1_0_spec: Path, caching: bool = False
    ) -> None:
        self.test_run_conformance(cwl_v1_0_spec, batchSystem="lsf", caching=caching)

    @slow
    @needs_slurm
    @pytest.mark.slow
    @pytest.mark.slurm
    @pytest.mark.skip("Not run")
    def test_slurm_cwl_conformance(
        self, cwl_v1_0_spec: Path, caching: bool = False
    ) -> None:
        self.test_run_conformance(cwl_v1_0_spec, batchSystem="slurm", caching=caching)

    @slow
    @needs_torque
    @pytest.mark.slow
    @pytest.mark.torque
    @pytest.mark.skip("Not run")
    def test_torque_cwl_conformance(
        self, cwl_v1_0_spec: Path, caching: bool = False
    ) -> None:
        self.test_run_conformance(cwl_v1_0_spec, batchSystem="torque", caching=caching)

    @slow
    @needs_gridengine
    @pytest.mark.slow
    @pytest.mark.gridengine
    @pytest.mark.skip("Not run")
    def test_gridengine_cwl_conformance(
        self, cwl_v1_0_spec: Path, caching: bool = False
    ) -> None:
        self.test_run_conformance(
            cwl_v1_0_spec, batchSystem="grid_engine", caching=caching
        )

    @slow
    @needs_mesos
    @pytest.mark.slow
    @pytest.mark.mesos
    @pytest.mark.skip("Not run")
    def test_mesos_cwl_conformance(
        self, cwl_v1_0_spec: Path, caching: bool = False
    ) -> None:
        self.test_run_conformance(cwl_v1_0_spec, batchSystem="mesos", caching=caching)

    @slow
    @needs_kubernetes
    @pytest.mark.slow
    @pytest.mark.kubernetes
    @pytest.mark.online
    def test_kubernetes_cwl_conformance(
        self, cwl_v1_0_spec: Path, caching: bool = False
    ) -> None:
        self.test_run_conformance(
            cwl_v1_0_spec,
            caching=caching,
            batchSystem="kubernetes",
            extra_args=["--retryCount=3"],
            # This test doesn't work with
            # Singularity; see
            # https://github.com/common-workflow-language/cwltool/blob/7094ede917c2d5b16d11f9231fe0c05260b51be6/conformance-test.sh#L99-L117
            skipped_tests="docker_entrypoint",
        )

    @slow
    @needs_lsf
    @pytest.mark.slow
    @pytest.mark.lsf
    @pytest.mark.skip(reason="Not run")
    def test_lsf_cwl_conformance_with_caching(self, cwl_v1_0_spec: Path) -> None:
        self.test_lsf_cwl_conformance(cwl_v1_0_spec, caching=True)

    @slow
    @needs_slurm
    @pytest.mark.slow
    @pytest.mark.slurm
    @pytest.mark.skip(reason="Not run")
    def test_slurm_cwl_conformance_with_caching(self, cwl_v1_0_spec: Path) -> None:
        self.test_slurm_cwl_conformance(cwl_v1_0_spec, caching=True)

    @slow
    @needs_torque
    @pytest.mark.slow
    @pytest.mark.torque
    @pytest.mark.skip(reason="Not run")
    def test_torque_cwl_conformance_with_caching(self, cwl_v1_0_spec: Path) -> None:
        self.test_torque_cwl_conformance(cwl_v1_0_spec, caching=True)

    @slow
    @needs_gridengine
    @pytest.mark.slow
    @pytest.mark.gridengine
    @pytest.mark.skip(reason="Not run")
    def test_gridengine_cwl_conformance_with_caching(self, cwl_v1_0_spec: Path) -> None:
        self.test_gridengine_cwl_conformance(cwl_v1_0_spec, caching=True)

    @slow
    @needs_mesos
    @pytest.mark.slow
    @pytest.mark.mesos
    @pytest.mark.skip(reason="Not run")
    def test_mesos_cwl_conformance_with_caching(self, cwl_v1_0_spec: Path) -> None:
        self.test_mesos_cwl_conformance(cwl_v1_0_spec, caching=True)

    @slow
    @needs_kubernetes
    @pytest.mark.slow
    @pytest.mark.kubernetes
    @pytest.mark.online
    def test_kubernetes_cwl_conformance_with_caching(self, cwl_v1_0_spec: Path) -> None:
        self.test_kubernetes_cwl_conformance(cwl_v1_0_spec, caching=True)


@pytest.fixture(scope="function")
def cwl_v1_1_spec(tmp_path: Path) -> Generator[Path]:
    # The latest cwl git commit hash from https://github.com/common-workflow-language/cwl-v1.1
    # Update it to get the latest tests.
    testhash = "664835e83eb5e57eee18a04ce7b05fb9d70d77b7"
    url = (
        "https://github.com/common-workflow-language/cwl-v1.1/archive/%s.zip" % testhash
    )
    urlretrieve(url, "spec.zip")
    with zipfile.ZipFile("spec.zip", "r") as z:
        z.extractall()
    shutil.move("cwl-v1.1-%s" % testhash, str(tmp_path))
    os.remove("spec.zip")
    try:
        yield tmp_path / ("cwl-v1.1-%s" % testhash)
    finally:
        pass  # no cleanup


@needs_cwl
@needs_online
@pytest.mark.cwl
@pytest.mark.online
class TestCWLv11:
    """
    Run the CWL 1.1 conformance tests in various environments.
    """

    @slow
    @needs_docker
    @pytest.mark.slow
    @pytest.mark.docker
    @pytest.mark.online
    @pytest.mark.timeout(CONFORMANCE_TEST_TIMEOUT)
    def test_run_conformance(
        self,
        cwl_v1_1_spec: Path,
        caching: bool = False,
        batchSystem: Optional[str] = None,
        skipped_tests: Optional[str] = None,
        extra_args: Optional[list[str]] = None,
    ) -> None:
        run_conformance_tests(
            workDir=str(cwl_v1_1_spec),
            yml=str(cwl_v1_1_spec / "conformance_tests.yaml"),
            caching=caching,
            batchSystem=batchSystem,
            skipped_tests=skipped_tests,
            extra_args=extra_args,
        )

    @slow
    @needs_docker
    @pytest.mark.slow
    @pytest.mark.docker
    @pytest.mark.online
    @pytest.mark.timeout(CONFORMANCE_TEST_TIMEOUT)
    def test_run_conformance_with_caching(self, cwl_v1_1_spec: Path) -> None:
        self.test_run_conformance(cwl_v1_1_spec, caching=True)

    @slow
    @needs_kubernetes
    @pytest.mark.slow
    @pytest.mark.kubernetes
    @pytest.mark.online
    def test_kubernetes_cwl_conformance(
        self, cwl_v1_1_spec: Path, caching: bool = False
    ) -> None:
        self.test_run_conformance(
            cwl_v1_1_spec,
            batchSystem="kubernetes",
            extra_args=["--retryCount=3"],
            # These tests don't work with
            # Singularity; see
            # https://github.com/common-workflow-language/cwltool/blob/7094ede917c2d5b16d11f9231fe0c05260b51be6/conformance-test.sh#L99-L117
            skipped_tests="docker_entrypoint,stdin_shorcut",
            caching=caching,
        )

    @slow
    @needs_kubernetes
    @pytest.mark.slow
    @pytest.mark.kubernetes
    @pytest.mark.online
    def test_kubernetes_cwl_conformance_with_caching(self, cwl_v1_1_spec: Path) -> None:
        self.test_kubernetes_cwl_conformance(cwl_v1_1_spec, caching=True)


@pytest.fixture(scope="function")
def cwl_v1_2_spec(tmp_path: Path) -> Generator[Path]:
    # The latest cwl git commit hash from https://github.com/common-workflow-language/cwl-v1.2
    # Update it to get the latest tests.
    testhash = "0d538a0dbc5518f3c6083ce4571926f65cb84f76"
    url = (
        "https://github.com/common-workflow-language/cwl-v1.2/archive/%s.zip" % testhash
    )
    urlretrieve(url, "spec.zip")
    with zipfile.ZipFile("spec.zip", "r") as z:
        z.extractall()
    shutil.move("cwl-v1.2-%s" % testhash, str(tmp_path))
    os.remove("spec.zip")
    try:
        yield tmp_path / ("cwl-v1.2-%s" % testhash)
    finally:
        pass  # no cleanup


@needs_cwl
@needs_online
@pytest.mark.cwl
@pytest.mark.online
class TestCWLv12:
    """
    Run the CWL 1.2 conformance tests in various environments.
    """

    @slow
    @needs_docker
    @pytest.mark.slow
    @pytest.mark.docker
    @pytest.mark.online
    @pytest.mark.timeout(CONFORMANCE_TEST_TIMEOUT)
    def test_run_conformance(
        self,
        cwl_v1_2_spec: Path,
        runner: Optional[str] = None,
        caching: bool = False,
        batchSystem: Optional[str] = None,
        selected_tests: Optional[str] = None,
        skipped_tests: Optional[str] = None,
        extra_args: Optional[list[str]] = None,
        must_support_all_features: bool = False,
        junit_file: Optional[str] = None,
    ) -> None:
        if junit_file is None:
            junit_file = os.path.abspath("conformance-1.2.junit.xml")
        run_conformance_tests(
            workDir=str(cwl_v1_2_spec),
            yml=str(cwl_v1_2_spec / "conformance_tests.yaml"),
            runner=runner,
            caching=caching,
            batchSystem=batchSystem,
            selected_tests=selected_tests,
            skipped_tests=skipped_tests,
            extra_args=extra_args,
            must_support_all_features=must_support_all_features,
            junit_file=junit_file,
        )

    @slow
    @needs_docker
    @pytest.mark.slow
    @pytest.mark.docker
    @pytest.mark.online
    @pytest.mark.timeout(CONFORMANCE_TEST_TIMEOUT)
    def test_run_conformance_with_caching(self, cwl_v1_2_spec: Path) -> None:
        self.test_run_conformance(
            cwl_v1_2_spec,
            caching=True,
            junit_file=os.path.abspath("caching-conformance-1.2.junit.xml"),
        )

    @slow
    @needs_docker
    @pytest.mark.slow
    @pytest.mark.docker
    @pytest.mark.timeout(CONFORMANCE_TEST_TIMEOUT)
    def test_run_conformance_with_task_caching(
        self, cwl_v1_2_spec: Path, tmp_path: Path
    ) -> None:
        self.test_run_conformance(
            cwl_v1_2_spec,
            junit_file=os.path.abspath("task-caching-conformance-1.2.junit.xml"),
            extra_args=["--cachedir", str(tmp_path / "task_cache")],
        )

    @slow
    @needs_docker
    @pytest.mark.slow
    @pytest.mark.docker
    @pytest.mark.timeout(CONFORMANCE_TEST_TIMEOUT)
    def test_run_conformance_with_in_place_update(self, cwl_v1_2_spec: Path) -> None:
        """
        Make sure that with --bypass-file-store we properly support in place
        update on a single node, and that this doesn't break any other
        features.
        """
        self.test_run_conformance(
            cwl_v1_2_spec,
            extra_args=["--bypass-file-store"],
            must_support_all_features=True,
            junit_file=os.path.abspath("in-place-update-conformance-1.2.junit.xml"),
        )

    @slow
    @needs_kubernetes
    @pytest.mark.slow
    @pytest.mark.kubernetes
    @pytest.mark.online
    def test_kubernetes_cwl_conformance(
        self,
        cwl_v1_2_spec: Path,
        caching: bool = False,
        junit_file: Optional[str] = None,
    ) -> None:
        if junit_file is None:
            junit_file = os.path.abspath("kubernetes-conformance-1.2.junit.xml")
        self.test_run_conformance(
            cwl_v1_2_spec,
            caching=caching,
            batchSystem="kubernetes",
            extra_args=["--retryCount=3"],
            # This test doesn't work with
            # Singularity; see
            # https://github.com/common-workflow-language/cwltool/blob/7094ede917c2d5b16d11f9231fe0c05260b51be6/conformance-test.sh#L99-L117
            # and
            # https://github.com/common-workflow-language/cwltool/issues/1441#issuecomment-826747975
            skipped_tests="docker_entrypoint",
            junit_file=junit_file,
        )

    @slow
    @needs_kubernetes
    @pytest.mark.slow
    @pytest.mark.kubernetes
    @pytest.mark.online
    def test_kubernetes_cwl_conformance_with_caching(self, cwl_v1_2_spec: Path) -> None:
        self.test_kubernetes_cwl_conformance(
            cwl_v1_2_spec,
            caching=True,
            junit_file=os.path.abspath("kubernetes-caching-conformance-1.2.junit.xml"),
        )

    @slow
    @needs_wes_server
    @pytest.mark.slow
    @pytest.mark.wes_server
    @pytest.mark.online
    def test_wes_server_cwl_conformance(self, cwl_v1_2_spec: Path) -> None:
        """
        Run the CWL conformance tests via WES. TOIL_WES_ENDPOINT must be
        specified. If the WES server requires authentication, set TOIL_WES_USER
        and TOIL_WES_PASSWORD.

        To run manually:

        TOIL_WES_ENDPOINT=http://localhost:8080 \
        TOIL_WES_USER=test \
        TOIL_WES_PASSWORD=password \
        python -m pytest src/toil/test/cwl/cwlTest.py::TestCWLv12::test_wes_server_cwl_conformance -vv --log-level INFO --log-cli-level INFO
        """
        endpoint = os.environ.get("TOIL_WES_ENDPOINT")
        extra_args = [f"--wes_endpoint={endpoint}"]

        # These are the ones that currently fail:
        #   - 310: mixed_version_v10_wf
        #   - 311: mixed_version_v11_wf
        #   - 312: mixed_version_v12_wf

        # Main issues:
        # 1. `cwltool --print-deps` doesn't seem to include secondary files from the default
        #     e.g.: https://github.com/common-workflow-language/cwl-v1.2/blob/1.2.1_proposed/tests/mixed-versions/wf-v10.cwl#L4-L10

        self.test_run_conformance(
            cwl_v1_2_spec,
            runner="toil-wes-cwl-runner",
            selected_tests="1-309,313-337",
            extra_args=extra_args,
        )


@needs_cwl
@pytest.mark.cwl
@pytest.mark.cwl_small_log_dir
def test_workflow_echo_string_scatter_stderr_log_dir(tmp_path: Path) -> None:
    log_dir = tmp_path / "cwl-logs"
    with get_data("test/cwl/echo_string_scatter_capture_stdout.cwl") as cwl_file:
        cmd = [
            "toil-cwl-runner",
            f"--jobStore={tmp_path / 'jobstore'}",
            "--strict-memory-limit",
            f"--log-dir={log_dir}",
            str(cwl_file),
        ]
        p = subprocess.run(cmd, capture_output=True, text=True)
        outputs = json.loads(p.stdout)
        out_list = outputs["list_out"]
        assert len(out_list) == 2, f"outList shoud have two file elements {out_list}"
        out_base = outputs["list_out"][0]
        # This is a test on the scatter functionality and stdout.
        # Each value of scatter should generate a separate file in the output.
        for index, file in enumerate(out_list):
            if index > 0:
                new_file_loc = out_base["location"] + f"_{index + 1}"
            else:
                new_file_loc = out_base["location"]
            assert (
                new_file_loc == file["location"]
            ), f"Toil should have detected conflicts for these stdout files {new_file_loc} and {file}"

        assert "Finished toil run successfully" in p.stderr
        assert p.returncode == 0

        assert log_dir.exists()
        scatter_0 = log_dir / "echo-test-scatter.0.scatter"
        scatter_1 = log_dir / "echo-test-scatter.1.scatter"
        list_0 = log_dir / "echo-test-scatter.0.list"
        list_1 = log_dir / "echo-test-scatter.1.list"
        assert scatter_0.exists()
        assert scatter_1.exists()
        assert list_0.exists()
        assert list_1.exists()


@needs_cwl
@pytest.mark.cwl
@pytest.mark.cwl_small_log_dir
def test_log_dir_echo_no_output(tmp_path: Path) -> None:
    log_dir = tmp_path / "cwl-logs"
    job_store = tmp_path / "test_log_dir_echo_no_output"
    with get_data("test/cwl/echo-stdout-log-dir.cwl") as cwl_file:
        cmd = [
            "toil-cwl-runner",
            f"--jobStore={job_store}",
            "--strict-memory-limit",
            f"--log-dir={str(log_dir)}",
            str(cwl_file),
        ]
        subprocess.run(cmd)

        assert log_dir.exists()
        assert sum(1 for _ in log_dir.iterdir()) == 1

        subdir = next(log_dir.iterdir())
        assert subdir.name == "echo"
        assert subdir.is_dir()
        assert sum(1 for _ in subdir.iterdir()) == 1
        result = next(subdir.iterdir())
        assert result.name == "out.txt"
        assert "hello" in result.read_text()


@needs_cwl
@pytest.mark.cwl
@pytest.mark.cwl_small_log_dir
def test_log_dir_echo_stderr(tmp_path: Path) -> None:
    log_dir = tmp_path / "cwl-logs"
    log_dir.mkdir()
    with get_data("test/cwl/echo-stderr.cwl") as cwl_file:
        cmd = [
            "toil-cwl-runner",
            f"--jobStore={str(tmp_path / 'test_log_dir_echo_stderr')}",
            "--strict-memory-limit",
            "--force-docker-pull",
            "--clean=always",
            f"--log-dir={str(log_dir)}",
            str(cwl_file),
        ]
        subprocess.run(cmd)
        tmp_path = log_dir

        assert len(list(tmp_path.iterdir())) == 1

        subdir = next(tmp_path.iterdir())
        assert subdir.name == "echo-stderr.cwl"
        assert subdir.is_dir()
        assert len(list(subdir.iterdir())) == 1
        result = next(subdir.iterdir())
        assert result.name == "out.txt"
        output = open(result).read()
        assert output == "hello\n"


# TODO: It's not clear how this test tests filename conflict resolution; it
# seems like it runs a python script to copy some files and makes sure the
# workflow doesn't fail.
@needs_cwl
@pytest.mark.cwl
@pytest.mark.cwl_small_log_dir
def test_filename_conflict_resolution(tmp_path: Path) -> None:
    with get_data("test/cwl/test_filename_conflict_resolution.cwl") as cwl_file:
        with get_data("test/cwl/test_filename_conflict_resolution.ms") as msin:
            cmd = [
                "toil-cwl-runner",
                f"--outdir={tmp_path}",
                str(cwl_file),
                "--msin",
                str(msin),
            ]
            p = subprocess.run(cmd, capture_output=True, text=True)
            assert "Finished toil run successfully" in p.stderr
            assert p.returncode == 0


@needs_cwl
@pytest.mark.cwl
@pytest.mark.cwl_small_log_dir
def test_filename_conflict_resolution_3_or_more(tmp_path: Path) -> None:
    with get_data("test/cwl/scatter_duplicate_outputs.cwl") as cwl_file:
        cmd = ["toil-cwl-runner", f"--outdir={tmp_path}", str(cwl_file)]
        p = subprocess.run(cmd, capture_output=True, text=True)
        assert "Finished toil run successfully" in p.stderr
        assert p.returncode == 0
        assert (
            sum(1 for _ in tmp_path.iterdir()) == 9
        ), f"All 9 files made by the scatter should be in the directory: {tmp_path}"


@needs_cwl
@needs_docker
@pytest.mark.cwl
@pytest.mark.docker
@pytest.mark.cwl_small_log_dir
def test_filename_conflict_detection(tmp_path: Path) -> None:
    """
    Make sure we don't just stage files over each other when using a container.
    """
    with get_data("test/cwl/test_filename_conflict_detection.cwl") as cwl_file:
        cmd = ["toil-cwl-runner", f"--outdir={tmp_path}", str(cwl_file)]
        p = subprocess.run(cmd, capture_output=True, text=True)
        assert "File staging conflict" in p.stderr
        assert p.returncode != 0


@needs_cwl
@needs_docker
@pytest.mark.cwl
@pytest.mark.docker
@pytest.mark.cwl_small_log_dir
def test_filename_conflict_detection_at_root(tmp_path: Path) -> None:
    """
    Make sure we don't just stage files over each other.

    Specifically, when using a container and the files are at the root of the work dir.
    """
    with get_data("test/cwl/test_filename_conflict_detection_at_root.cwl") as cwl_file:
        cmd = ["toil-cwl-runner", f"--outdir={tmp_path}", str(cwl_file)]
        p = subprocess.run(cmd, capture_output=True, text=True)
        assert "File staging conflict" in p.stderr
        assert p.returncode != 0


@needs_cwl
@pytest.mark.cwl
@pytest.mark.cwl_small
def test_pick_value_with_one_null_value(
    caplog: pytest.LogCaptureFixture, tmp_path: Path
) -> None:
    """
    Make sure toil-cwl-runner does not false log a warning when pickValue is
    used but outputSource only contains one null value. See: #3991.
    """
    from toil.cwl import cwltoil

    with get_data("test/cwl/conditional_wf.cwl") as cwl_file:
        with get_data("test/cwl/conditional_wf.yaml") as job_file:
            with caplog.at_level(logging.WARNING, logger="toil.cwl.cwltoil"):
                cwltoil.main([f"--outdir={tmp_path}", str(cwl_file), str(job_file)])
                for line in caplog.messages:
                    assert (
                        "You had a conditional step that did not run, but you did not use pickValue to handle the skipped input."
                        not in line
                    )


@needs_cwl
@pytest.mark.cwl
@pytest.mark.cwl_small
def test_workflow_echo_string(tmp_path: Path) -> None:
    with get_data("test/cwl/echo_string.cwl") as cwl_file:
        cmd = [
            "toil-cwl-runner",
            f"--jobStore=file:{tmp_path / 'jobstore'}",
            "--strict-memory-limit",
            str(cwl_file),
        ]
        p = subprocess.run(cmd, capture_output=True, text=True)
        assert (
            p.stdout.strip() == "{}"
        ), f"Got wrong output: {p.stdout}\nWith error: {p.stderr}"
        assert "Finished toil run successfully" in p.stderr
        assert p.returncode == 0


@needs_cwl
@pytest.mark.cwl
@pytest.mark.cwl_small
def test_workflow_echo_string_scatter_capture_stdout(tmp_path: Path) -> None:
    with get_data("test/cwl/echo_string_scatter_capture_stdout.cwl") as cwl_file:
        cmd = [
            "toil-cwl-runner",
            f"--jobStore=file:{tmp_path / 'jobStore'}",
            "--strict-memory-limit",
            str(cwl_file),
        ]
        p = subprocess.run(cmd, capture_output=True, text=True)
        assert len(p.stdout) > 0
        outputs = json.loads(p.stdout)
        out_list = outputs["list_out"]
        assert len(out_list) == 2, f"outList shoud have two file elements {out_list}"
        out_base = outputs["list_out"][0]
        # This is a test on the scatter functionality and stdout.
        # Each value of scatter should generate a separate file in the output.
        for index, file in enumerate(out_list):
            if index > 0:
                new_file_loc = out_base["location"] + f"_{index + 1}"
            else:
                new_file_loc = out_base["location"]
            assert (
                new_file_loc == file["location"]
            ), f"Toil should have detected conflicts for these stdout files {new_file_loc} and {file}"

        assert "Finished toil run successfully" in p.stderr
        assert p.returncode == 0


@needs_cwl
@pytest.mark.cwl
@pytest.mark.cwl_small
def test_visit_top_cwl_class() -> None:
    structure = {
        "class": "Directory",
        "listing": [
            {
                "class": "Directory",
                "listing": [
                    {"class": "File"},
                    {
                        "class": "File",
                        "secondaryFiles": [
                            {"class": "Directory"},
                            {"class": "File"},
                            {"cruft"},
                        ],
                    },
                ],
            },
            {"some garbage": "yep"},
            [],
            None,
        ],
    }

    counter = 0

    def increment(thing: "CWLObjectType") -> None:
        """
        Make sure we are at something CWL object like, and count it.
        """
        assert "class" in thing
        nonlocal counter
        counter += 1

    # We should stop at the root when looking for a Directory
    visit_top_cwl_class(structure, ("Directory",), increment)
    assert counter == 1

    # We should see the top-level files when looking for a file
    counter = 0
    visit_top_cwl_class(structure, ("File",), increment)
    assert counter == 2

    # When looking for a file or a directory we should stop at the first match to either.
    counter = 0
    visit_top_cwl_class(structure, ("File", "Directory"), increment)
    assert counter == 1


@needs_cwl
@pytest.mark.cwl
@pytest.mark.cwl_small
def test_visit_cwl_class_and_reduce() -> None:
    structure = {
        "class": "Directory",
        "listing": [
            {
                "class": "Directory",
                "listing": [
                    {"class": "File"},
                    {
                        "class": "File",
                        "secondaryFiles": [
                            {"class": "Directory"},
                            {"class": "File"},
                            {"cruft"},
                        ],
                    },
                ],
            },
            {"some garbage": "yep"},
            [],
            None,
        ],
    }

    down_count = 0

    def op_down(thing: "CWLObjectType") -> int:
        """
        Grab the ID of the thing we are at, and count what we visit going
        down.
        """
        nonlocal down_count
        down_count += 1
        return id(thing)

    up_count = 0
    up_child_count = 0

    def op_up(thing: "CWLObjectType", down_value: int, child_results: list[str]) -> str:
        """
        Check the down return value and the up return values, and count
        what we visit going up and what child relationships we have.
        """
        nonlocal up_child_count
        nonlocal up_count
        assert down_value == id(thing)
        for res in child_results:
            assert res == "Sentinel value!"
            up_child_count += 1
        up_count += 1
        return "Sentinel value!"

    visit_cwl_class_and_reduce(structure, ("Directory",), op_down, op_up)
    assert down_count == 3
    assert up_count == 3
    # Only 2 child relationships
    assert up_child_count == 2


@needs_cwl
@pytest.mark.cwl
@pytest.mark.cwl_small
def test_download_structure(tmp_path: Path) -> None:
    """
    Make sure that download_structure makes the right calls to what it thinks is the file store.
    """

    # Define what we would download
    fid1 = FileID("afile", 10, False)
    fid2 = FileID("adifferentfile", 1000, True)

    # And what directory structure it would be in
    structure: DirectoryStructure = {
        "dir1": {
            "dir2": {
                "f1": "toilfile:" + fid1.pack(),
                "f1again": "toilfile:" + fid1.pack(),
                "dir2sub": {},
            },
            "dir3": {},
        },
        "anotherfile": "toilfile:" + fid2.pack(),
    }

    # Say where to put it on the filesystem
    to_dir = tmp_path

    # Make a fake file store
    file_store = Mock(AbstractFileStore)

    # These will be populated.
    # TODO: This cache seems unused. Remove it?
    # This maps filesystem path to CWL URI
    index: dict[str, str] = {}
    # This maps CWL URI to filesystem path
    existing: dict[str, str] = {}

    # Do the download
    download_structure(file_store, index, existing, structure, str(to_dir))

    # Check the results
    # 3 files should be made
    assert len(index) == 3
    # From 2 unique URIs
    assert len(existing) == 2

    # Make sure that the index contents (path to URI) are correct
    assert str(to_dir / "dir1/dir2/f1") in index
    assert str(to_dir / "dir1/dir2/f1again") in index
    assert str(to_dir / "anotherfile") in index
    assert (
        index[str(to_dir / "dir1/dir2/f1")]
        == cast(
            DirectoryStructure, cast(DirectoryStructure, structure["dir1"])["dir2"]
        )["f1"]
    )
    assert (
        index[str(to_dir / "dir1/dir2/f1again")]
        == cast(
            DirectoryStructure, cast(DirectoryStructure, structure["dir1"])["dir2"]
        )["f1again"]
    )
    assert index[str(to_dir / "anotherfile")] == structure["anotherfile"]

    # And the existing contents (URI to path)
    assert "toilfile:" + fid1.pack() in existing
    assert "toilfile:" + fid2.pack() in existing
    assert existing["toilfile:" + fid1.pack()] in [
        str(to_dir / "dir1/dir2/f1"),
        str(to_dir / "dir1/dir2/f1again"),
    ]
    assert existing["toilfile:" + fid2.pack()] == str(to_dir / "anotherfile")

    # The directory structure should be created for real
    assert (to_dir / "dir1").is_dir()
    assert (to_dir / "dir1/dir2").is_dir()
    assert (to_dir / "dir1/dir2/dir2sub").is_dir()
    assert (to_dir / "dir1/dir3").is_dir()

    # The file store should have been asked to do the download
    file_store.readGlobalFile.assert_has_calls(
        [
            call(fid1, str(to_dir / "dir1/dir2/f1"), symlink=False),
            call(fid1, str(to_dir / "dir1/dir2/f1again"), symlink=False),
            call(fid2, str(to_dir / "anotherfile"), symlink=False),
        ],
        any_order=True,
    )


@needs_cwl
@pytest.mark.cwl
@pytest.mark.timeout(300)
def test_import_on_workers() -> None:
    from toil.cwl import cwltoil

    detector = ImportWorkersMessageHandler()

    # Set up a log message detector to the root logger
    logging.getLogger().addHandler(detector)

    with get_data("test/cwl/download.cwl") as cwl_file:
        with get_data("test/cwl/directory/directory/file.txt") as file_path:
            args = [
                "--runImportsOnWorkers",
                "--importWorkersDisk=10MiB",
                "--realTimeLogging=True",
                "--logLevel=INFO",
                "--logColors=False",
                str(cwl_file),
                "--input",
                str(file_path),
            ]
            cwltoil.main(args)

        assert detector.detected is True


# StreamHandler is generic, _typeshed doesn't exist at runtime, do a bit of typing trickery, see https://github.com/python/typeshed/issues/5680
if TYPE_CHECKING:
    from _typeshed import SupportsWrite

    _stream_handler = logging.StreamHandler[SupportsWrite[str]]
else:
    _stream_handler = logging.StreamHandler


class ImportWorkersMessageHandler(_stream_handler):
    """
    Detect the import workers log message and set a flag.
    """

    def __init__(self) -> None:
        self.detected = False  # Have we seen the message we want?

        super().__init__(sys.stderr)

    def emit(self, record: logging.LogRecord) -> None:
        if (record.msg % record.args).startswith(
            "Issued job 'CWLImportJob' CWLImportJob"
        ):
            self.detected = True



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import os
import subprocess
import sys

env_var = 'GITLAB_SECRET_FILE_QUAY_CREDENTIALS'

if env_var not in os.environ:
    print('Error: could not find environment variable ' + env_var)
    sys.exit(1)

print('Starting quay.io login process...')
with open(os.environ[env_var], 'r') as f:
    keys = json.load(f)
process = subprocess.Popen('docker login quay.io -u "{user}" --password-stdin'.format(user=keys['user']),
                           stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE,
                           shell=True)
print('Logging into quay.io...')
stdout, stderr = process.communicate(input=keys['password'].encode('utf-8'))

if 'Login Succeeded' in str(stdout):
    print('Login Succeeded.')
else:
    print('Error while attempting to log into quay.io:\n' + str(stderr))
    sys.exit(1)



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import configparser
import json
import logging
import os.path
import subprocess
import tempfile
import textwrap
from abc import ABC, abstractmethod
from functools import total_ordering
from typing import Any, Optional, Union
from urllib.parse import quote
from uuid import uuid4

from toil import applianceSelf, customDockerInitCmd, customInitCmd
from toil.provisioners import ClusterTypeNotSupportedException
from toil.provisioners.node import Node

a_short_time = 5
logger = logging.getLogger(__name__)


class ManagedNodesNotSupportedException(RuntimeError):
    """
    Raised when attempting to add managed nodes (which autoscale up and down by
    themselves, without the provisioner doing the work) to a provisioner that
    does not support them.

    Polling with this and try/except is the Right Way to check if managed nodes
    are available from a provisioner.
    """


@total_ordering
class Shape:
    """
    Represents a job or a node's "shape", in terms of the dimensions of memory, cores, disk and
    wall-time allocation.

    The wallTime attribute stores the number of seconds of a node allocation, e.g. 3600 for AWS.
    FIXME: and for jobs?

    The memory and disk attributes store the number of bytes required by a job (or provided by a
    node) in RAM or on disk (SSD or HDD), respectively.
    """

    def __init__(
        self,
        wallTime: Union[int, float],
        memory: int,
        cores: Union[int, float],
        disk: int,
        preemptible: bool,
    ) -> None:
        self.wallTime = wallTime
        self.memory = memory
        self.cores = cores
        self.disk = disk
        self.preemptible = preemptible

    def __eq__(self, other: Any) -> bool:
        return (
            self.wallTime == other.wallTime
            and self.memory == other.memory
            and self.cores == other.cores
            and self.disk == other.disk
            and self.preemptible == other.preemptible
        )

    def greater_than(self, other: Any) -> bool:
        if self.preemptible < other.preemptible:
            return True
        elif self.preemptible > other.preemptible:
            return False
        elif self.memory > other.memory:
            return True
        elif self.memory < other.memory:
            return False
        elif self.cores > other.cores:
            return True
        elif self.cores < other.cores:
            return False
        elif self.disk > other.disk:
            return True
        elif self.disk < other.disk:
            return False
        elif self.wallTime > other.wallTime:
            return True
        elif self.wallTime < other.wallTime:
            return False
        else:
            return False

    def __gt__(self, other: Any) -> bool:
        return self.greater_than(other)

    def __repr__(self) -> str:
        return "Shape(wallTime=%s, memory=%s, cores=%s, disk=%s, preemptible=%s)" % (
            self.wallTime,
            self.memory,
            self.cores,
            self.disk,
            self.preemptible,
        )

    def __str__(self) -> str:
        return self.__repr__()

    def __hash__(self) -> int:
        # Since we replaced __eq__ we need to replace __hash__ as well.
        return hash(
            (self.wallTime, self.memory, self.cores, self.disk, self.preemptible)
        )


class AbstractProvisioner(ABC):
    """Interface for provisioning worker nodes to use in a Toil cluster."""

    LEADER_HOME_DIR = "/root/"  # home directory in the Toil appliance on an instance
    cloud: str = None

    def __init__(
        self,
        clusterName: Optional[str] = None,
        clusterType: Optional[str] = "mesos",
        zone: Optional[str] = None,
        nodeStorage: int = 50,
        nodeStorageOverrides: Optional[list[str]] = None,
        enable_fuse: bool = False,
    ) -> None:
        """
        Initialize provisioner.

        Implementations should raise ClusterTypeNotSupportedException if
        presented with an unimplemented clusterType.

        :param clusterName: The cluster identifier.
        :param clusterType: The kind of cluster to make; 'mesos' or 'kubernetes'.
        :param zone: The zone the cluster runs in.
        :param nodeStorage: The amount of storage on the worker instances, in gigabytes.
        """
        self.clusterName = clusterName
        self.clusterType = clusterType

        if self.clusterType not in self.supportedClusterTypes():
            # This isn't actually a cluster type we can do
            raise ClusterTypeNotSupportedException(type(self), clusterType)

        self._zone = zone
        self._nodeStorage = nodeStorage
        self._nodeStorageOverrides = {}
        for override in nodeStorageOverrides or []:
            nodeShape, storageOverride = override.split(":")
            self._nodeStorageOverrides[nodeShape] = int(storageOverride)
        self._leaderPrivateIP: Optional[str] = None
        # This will hold an SSH public key for Mesos clusters, or the
        # Kubernetes joining information as a dict for Kubernetes clusters.
        self._leaderWorkerAuthentication = None

        # Whether or not to use FUSE on the cluster. If true, the cluster's Toil containers will be launched in privileged mode
        self.enable_fuse = enable_fuse

        if clusterName:
            # Making a new cluster
            self.createClusterSettings()
        else:
            # Starting up on an existing cluster
            self.readClusterSettings()

    @abstractmethod
    def supportedClusterTypes(self) -> set[str]:
        """
        Get all the cluster types that this provisioner implementation
        supports.
        """
        raise NotImplementedError

    @abstractmethod
    def createClusterSettings(self):
        """
        Initialize class for a new cluster, to be deployed, when running
        outside the cloud.
        """
        raise NotImplementedError

    @abstractmethod
    def readClusterSettings(self):
        """
        Initialize class from an existing cluster. This method assumes that
        the instance we are running on is the leader.

        Implementations must call _setLeaderWorkerAuthentication().
        """
        raise NotImplementedError

    def _write_file_to_cloud(self, key: str, contents: bytes) -> str:
        """
        Write a file to a physical storage system that is accessible to the
        leader and all nodes during the life of the cluster. Additional
        resources should be cleaned up in `self.destroyCluster()`.

        :return: A public URL that can be used to retrieve the file.
        """
        raise NotImplementedError

    def _read_file_from_cloud(self, key: str) -> bytes:
        """
        Return the contents of the file written by `self._write_file_to_cloud()`.
        """
        raise NotImplementedError

    def _get_user_data_limit(self) -> int:
        """
        Get the maximum number of bytes that can be passed as the user data
        during node creation.
        """
        raise NotImplementedError

    def _setLeaderWorkerAuthentication(self, leader: Node = None):
        """
        Configure authentication between the leader and the workers.

        Assumes that we are running on the leader, unless a Node is given, in
        which case credentials will be pulled from or created there.

        Configures the backing cluster scheduler so that the leader and workers
        will be able to communicate securely. Authentication may be one-way or
        mutual.

        Until this is called, new nodes may not be able to communicate with the
        leader. Afterward, the provisioner will include the necessary
        authentication information when provisioning nodes.

        :param leader: Node to pull credentials from, if not the current machine.
        """

        if self.clusterType == "mesos":
            # We're using a Mesos cluster, so set up SSH from leader to workers.
            self._leaderWorkerAuthentication = self._setSSH(leader=leader)
        elif self.clusterType == "kubernetes":
            # We're using a Kubernetes cluster.
            self._leaderWorkerAuthentication = self._getKubernetesJoiningInfo(
                leader=leader
            )

    def _clearLeaderWorkerAuthentication(self):
        """
        Forget any authentication information populated by
        _setLeaderWorkerAuthentication(). It will need to be called again to
        provision more workers.
        """

        self._leaderWorkerAuthentication = None

    def _setSSH(self, leader: Node = None) -> str:
        """
        Generate a key pair, save it in /root/.ssh/id_rsa.pub on the leader,
        and return the public key. The file /root/.sshSuccess is used to
        prevent this operation from running twice.

        Also starts the ssh agent on the local node, if operating on the local
        node.

        :param leader: Node to operate on, if not the current machine.

        :return: Public key, without the "ssh-rsa" part.
        """

        # To work locally or remotely we need to do all our setup work as one
        # big bash -c
        command = [
            "bash",
            "-c",
            (
                "set -e; if [ ! -e /root/.sshSuccess ] ; "
                'then ssh-keygen -f /root/.ssh/id_rsa -t rsa -N ""; '
                "touch /root/.sshSuccess; fi; chmod 700 /root/.ssh;"
            ),
        ]

        if leader is None:
            # Run locally
            subprocess.check_call(command)

            # Grab from local file
            with open("/root/.ssh/id_rsa.pub") as f:
                leaderPublicKey = f.read()
        else:
            # Run remotely
            leader.sshInstance(*command, appliance=True)

            # Grab from remote file
            with tempfile.TemporaryDirectory() as tmpdir:
                localFile = os.path.join(tmpdir, "id_rsa.pub")
                leader.extractFile("/root/.ssh/id_rsa.pub", localFile, "toil_leader")

                with open(localFile) as f:
                    leaderPublicKey = f.read()

        # Drop the key type and keep just the key data
        leaderPublicKey = leaderPublicKey.split(" ")[1]

        # confirm it really is an RSA public key
        assert leaderPublicKey.startswith("AAAAB3NzaC1yc2E"), leaderPublicKey
        return leaderPublicKey

    def _getKubernetesJoiningInfo(self, leader: Node = None) -> dict[str, str]:
        """
        Get the Kubernetes joining info created when Kubernetes was set up on
        this node, which is the leader, or on a different specified Node.

        Returns a dict of JOIN_TOKEN, JOIN_CERT_HASH, and JOIN_ENDPOINT, which
        can be inserted into our Kubernetes worker setup script and config.

        :param leader: Node to operate on, if not the current machine.
        """

        # Make a parser for the config
        config = configparser.ConfigParser(interpolation=None)
        # Leave case alone
        config.optionxform = str

        if leader is None:
            # This info is always supposed to be set up before the Toil appliance
            # starts, and mounted in at the same path as on the host. So we just go
            # read it.
            with open("/etc/kubernetes/worker.ini") as f:
                config.read_file(f)
        else:
            # Grab from remote file
            with tempfile.TemporaryDirectory() as tmpdir:
                localFile = os.path.join(tmpdir, "worker.ini")
                leader.extractFile(
                    "/etc/kubernetes/worker.ini", localFile, "toil_leader"
                )

                with open(localFile) as f:
                    config.read_file(f)

        # Grab everything out of the default section where our setup script put
        # it.
        return dict(config["DEFAULT"])

    def setAutoscaledNodeTypes(self, nodeTypes: list[tuple[set[str], Optional[float]]]):
        """
        Set node types, shapes and spot bids for Toil-managed autoscaling.
        :param nodeTypes: A list of node types, as parsed with parse_node_types.
        """
        # This maps from an equivalence class of instance names to a spot bid.
        self._spotBidsMap = {}

        # This maps from a node Shape object to the instance type that has that
        # shape. TODO: what if multiple instance types in a cloud provider have
        # the same shape (e.g. AMD and Intel instances)???
        self._shape_to_instance_type = {}

        for node_type in nodeTypes:
            preemptible = node_type[1] is not None
            if preemptible:
                # Record the spot bid for the whole equivalence class
                self._spotBidsMap[frozenset(node_type[0])] = node_type[1]
            for instance_type_name in node_type[0]:
                # Record the instance shape and associated type.
                shape = self.getNodeShape(instance_type_name, preemptible)
                self._shape_to_instance_type[shape] = instance_type_name

    def hasAutoscaledNodeTypes(self) -> bool:
        """
        Check if node types have been configured on the provisioner (via
        setAutoscaledNodeTypes).

        :returns: True if node types are configured for autoscaling, and false
                  otherwise.
        """
        return len(self.getAutoscaledInstanceShapes()) > 0

    def getAutoscaledInstanceShapes(self) -> dict[Shape, str]:
        """
        Get all the node shapes and their named instance types that the Toil
        autoscaler should manage.
        """

        if hasattr(self, "_shape_to_instance_type"):
            # We have had Toil-managed autoscaling set up
            return dict(self._shape_to_instance_type)
        else:
            # Nobody has called setAutoscaledNodeTypes yet, so nothing is to be autoscaled.
            return {}

    @staticmethod
    def retryPredicate(e):
        """
        Return true if the exception e should be retried by the cluster scaler.
        For example, should return true if the exception was due to exceeding an API rate limit.
        The error will be retried with exponential backoff.

        :param e: exception raised during execution of setNodeCount
        :return: boolean indicating whether the exception e should be retried
        """
        return False

    @abstractmethod
    def launchCluster(self, *args, **kwargs):
        """
        Initialize a cluster and create a leader node.

        Implementations must call _setLeaderWorkerAuthentication() with the
        leader so that workers can be launched.

        :param leaderNodeType: The leader instance.
        :param leaderStorage: The amount of disk to allocate to the leader in gigabytes.
        :param owner: Tag identifying the owner of the instances.

        """
        raise NotImplementedError

    @abstractmethod
    def addNodes(
        self,
        nodeTypes: set[str],
        numNodes: int,
        preemptible: bool,
        spotBid: Optional[float] = None,
    ) -> int:
        """
        Used to add worker nodes to the cluster

        :param numNodes: The number of nodes to add
        :param preemptible: whether or not the nodes will be preemptible
        :param spotBid: The bid for preemptible nodes if applicable (this can be set in config, also).
        :return: number of nodes successfully added
        """
        raise NotImplementedError

    def addManagedNodes(
        self, nodeTypes: set[str], minNodes, maxNodes, preemptible, spotBid=None
    ) -> None:
        """
        Add a group of managed nodes of the given type, up to the given maximum.
        The nodes will automatically be launched and terminated depending on cluster load.

        Raises ManagedNodesNotSupportedException if the provisioner
        implementation or cluster configuration can't have managed nodes.

        :param minNodes: The minimum number of nodes to scale to
        :param maxNodes: The maximum number of nodes to scale to
        :param preemptible: whether or not the nodes will be preemptible
        :param spotBid: The bid for preemptible nodes if applicable (this can be set in config, also).
        """

        # Not available by default
        raise ManagedNodesNotSupportedException(
            "Managed nodes not supported by this provisioner"
        )

    @abstractmethod
    def terminateNodes(self, nodes: list[Node]) -> None:
        """
        Terminate the nodes represented by given Node objects

        :param nodes: list of Node objects
        """
        raise NotImplementedError

    @abstractmethod
    def getLeader(self):
        """
        :return: The leader node.
        """
        raise NotImplementedError

    @abstractmethod
    def getProvisionedWorkers(
        self, instance_type: Optional[str] = None, preemptible: Optional[bool] = None
    ) -> list[Node]:
        """
        Gets all nodes, optionally of the given instance type or
        preemptability, from the provisioner. Includes both static and
        autoscaled nodes.

        :param preemptible: Boolean value to restrict to preemptible
               nodes or non-preemptible nodes
        :return: list of Node objects
        """
        raise NotImplementedError

    @abstractmethod
    def getNodeShape(self, instance_type: str, preemptible=False) -> Shape:
        """
        The shape of a preemptible or non-preemptible node managed by this provisioner. The node
        shape defines key properties of a machine, such as its number of cores or the time
        between billing intervals.

        :param str instance_type: Instance type name to return the shape of.
        """
        raise NotImplementedError

    @abstractmethod
    def destroyCluster(self) -> None:
        """
        Terminates all nodes in the specified cluster and cleans up all resources associated with the
        cluster.
        :param clusterName: identifier of the cluster to terminate.
        """
        raise NotImplementedError

    class InstanceConfiguration:
        """
        Allows defining the initial setup for an instance and then turning it
        into an Ignition configuration for instance user data.
        """

        def __init__(self):
            # Holds dicts with keys 'path', 'owner', 'permissions', and 'content' for files to create.
            # Permissions is a string octal number with leading 0.
            self.files = []
            # Holds dicts with keys 'name', 'command', and 'content' defining Systemd units to create
            self.units = []
            # Holds strings like "ssh-rsa actualKeyData" for keys to authorize (independently of cloud provider's system)
            self.sshPublicKeys = []

        def addFile(
            self,
            path: str,
            filesystem: str = "root",
            mode: Union[str, int] = "0755",
            contents: str = "",
            append: bool = False,
        ):
            """
            Make a file on the instance with the given filesystem, mode, and contents.

            See the storage.files section:
            https://github.com/kinvolk/ignition/blob/flatcar-master/doc/configuration-v2_2.md
            """
            if isinstance(mode, str):
                # Convert mode from octal to decimal
                mode = int(mode, 8)
            assert isinstance(mode, int)

            contents = "data:," + quote(contents.encode("utf-8"))

            ignition_file = {
                "path": path,
                "filesystem": filesystem,
                "mode": mode,
                "contents": {"source": contents},
            }

            if append:
                ignition_file["append"] = append

            self.files.append(ignition_file)

        def addUnit(self, name: str, enabled: bool = True, contents: str = ""):
            """
            Make a systemd unit on the instance with the given name (including
            .service), and content. Units will be enabled by default.

            Unit logs can be investigated with:
                systemctl status whatever.service
            or:
                journalctl -xe
            """

            self.units.append({"name": name, "enabled": enabled, "contents": contents})

        def addSSHRSAKey(self, keyData: str):
            """
            Authorize the given bare, encoded RSA key (without "ssh-rsa").
            """

            self.sshPublicKeys.append("ssh-rsa " + keyData)

        def toIgnitionConfig(self) -> str:
            """
            Return an Ignition configuration describing the desired config.
            """

            # Define the base config.  We're using Flatcar's v2.2.0 fork
            # See: https://github.com/kinvolk/ignition/blob/flatcar-master/doc/configuration-v2_2.md
            config = {
                "ignition": {"version": "2.2.0"},
                "storage": {"files": self.files},
                "systemd": {"units": self.units},
            }

            if len(self.sshPublicKeys) > 0:
                # Add SSH keys if needed
                config["passwd"] = {
                    "users": [{"name": "core", "sshAuthorizedKeys": self.sshPublicKeys}]
                }

            # Serialize as JSON
            return json.dumps(config, separators=(",", ":"))

    def getBaseInstanceConfiguration(self) -> InstanceConfiguration:
        """
        Get the base configuration for both leader and worker instances for all cluster types.
        """

        config = self.InstanceConfiguration()

        # We set Flatcar's update reboot strategy to off
        config.addFile(
            "/etc/coreos/update.conf",
            mode="0644",
            contents=textwrap.dedent(
                """\
        GROUP=stable
        REBOOT_STRATEGY=off
        """
            ),
        )

        # Then we have volume mounting. That always happens.
        self.addVolumesService(config)
        # We also always add the service to talk to Prometheus
        self.addNodeExporterService(config)

        return config

    def addVolumesService(self, config: InstanceConfiguration):
        """
        Add a service to prepare and mount local scratch volumes.
        """

        # TODO: when
        # https://www.flatcar.org/docs/latest/setup/storage/mounting-storage/
        # describes how to collect all the ephemeral disks declaratively and
        # make Ignition RAID them, stop doing it manually. Might depend on real
        # solution for https://github.com/coreos/ignition/issues/1126
        #
        # TODO: check what kind of instance this is, and what ephemeral volumes
        # *should* be there, and declaratively RAID and mount them.
        config.addFile(
            "/home/core/volumes.sh",
            contents=textwrap.dedent(
                """\
            #!/bin/bash
            set -x
            ephemeral_count=0
            drives=()
            # Directories are relative to /var
            directories=(lib/toil lib/mesos lib/docker lib/kubelet lib/cwl tmp)
            for drive in /dev/xvd{a..z} /dev/nvme{0..26}n1; do
                echo "checking for ${drive}"
                if [ -b $drive ]; then
                    echo "found it"
                    while [ "$(readlink -f "${drive}")" != "${drive}" ] ; do
                        drive="$(readlink -f "${drive}")"
                        echo "was a symlink to ${drive}"
                    done
                    seen=0
                    for other_drive in "${drives[@]}" ; do
                        if [ "${other_drive}" == "${drive}" ] ; then
                            seen=1
                            break
                        fi
                    done
                    if (( "${seen}" == "1" )) ; then
                        echo "already discovered via another name"
                        continue
                    fi
                    if mount | grep "^${drive}"; then
                        echo "already mounted, likely a root device"
                    else
                        ephemeral_count=$((ephemeral_count + 1 ))
                        drives+=("${drive}")
                        echo "increased ephemeral count by one"
                    fi
                fi
            done
            if (("$ephemeral_count" == "0" )); then
                echo "no ephemeral drive"
                for directory in "${directories[@]}"; do
                    sudo mkdir -p /var/$directory
                done
                exit 0
            fi
            sudo mkdir /mnt/ephemeral
            if (("$ephemeral_count" == "1" )); then
                echo "one ephemeral drive to mount"
                sudo mkfs.ext4 -F "${drives[@]}"
                sudo mount "${drives[@]}" /mnt/ephemeral
            fi
            if (("$ephemeral_count" > "1" )); then
                echo "multiple drives"
                for drive in "${drives[@]}"; do
                    sudo dd if=/dev/zero of=$drive bs=4096 count=1024
                done
                # determine force flag
                sudo mdadm --create -f --verbose /dev/md0 --level=0 --raid-devices=$ephemeral_count "${drives[@]}"
                sudo mkfs.ext4 -F /dev/md0
                sudo mount /dev/md0 /mnt/ephemeral
            fi
            for directory in "${directories[@]}"; do
                sudo mkdir -p /mnt/ephemeral/var/$directory
                sudo mkdir -p /var/$directory
                sudo mount --bind /mnt/ephemeral/var/$directory /var/$directory
            done
            """
            ),
        )
        # TODO: Make this retry?
        config.addUnit(
            "volume-mounting.service",
            contents=textwrap.dedent(
                """\
            [Unit]
            Description=mounts ephemeral volumes & bind mounts toil directories
            Before=docker.service

            [Service]
            Type=oneshot
            Restart=no
            ExecStart=/usr/bin/bash /home/core/volumes.sh

            [Install]
            WantedBy=multi-user.target
            """
            ),
        )

    def addNodeExporterService(self, config: InstanceConfiguration):
        """
        Add the node exporter service for Prometheus to an instance configuration.
        """

        config.addUnit(
            "node-exporter.service",
            contents=textwrap.dedent(
                """\
            [Unit]
            Description=node-exporter container
            After=docker.service

            [Service]
            Restart=on-failure
            RestartSec=2
            ExecStartPre=-/usr/bin/docker rm node_exporter
            ExecStart=/usr/bin/docker run \\
                -p 9100:9100 \\
                -v /proc:/host/proc \\
                -v /sys:/host/sys \\
                -v /:/rootfs \\
                --name node-exporter \\
                --restart always \\
                quay.io/prometheus/node-exporter:v1.3.1 \\
                --path.procfs /host/proc \\
                --path.sysfs /host/sys \\
                --collector.filesystem.ignored-mount-points ^/(sys|proc|dev|host|etc)($|/)

            [Install]
            WantedBy=multi-user.target
            """
            ),
        )

    def toil_service_env_options(self) -> str:
        return "-e TMPDIR=/var/tmp"

    def add_toil_service(
        self,
        config: InstanceConfiguration,
        role: str,
        keyPath: str = None,
        preemptible: bool = False,
    ):
        """
        Add the Toil leader or worker service to an instance configuration.

        Will run Mesos master or agent as appropriate in Mesos clusters.
        For Kubernetes clusters, will just sleep to provide a place to shell
        into on the leader, and shouldn't run on the worker.

        :param role: Should be 'leader' or 'worker'. Will not work for 'worker' until leader credentials have been collected.
        :param keyPath: path on the node to a server-side encryption key that will be added to the node after it starts. The service will wait until the key is present before starting.
        :param preemptible: Whether a worker should identify itself as preemptible or not to the scheduler.
        """

        # If keys are rsynced, then the mesos-agent needs to be started after the keys have been
        # transferred. The waitForKey.sh script loops on the new VM until it finds the keyPath file, then it starts the
        # mesos-agent. If there are multiple keys to be transferred, then the last one to be transferred must be
        # set to keyPath.
        MESOS_LOG_DIR = "--log_dir=/var/lib/mesos "
        LEADER_DOCKER_ARGS = "--registry=in_memory --cluster={name}"
        # --no-systemd_enable_support is necessary in Ubuntu 16.04 (otherwise,
        # Mesos attempts to contact systemd but can't find its run file)
        WORKER_DOCKER_ARGS = "--work_dir=/var/lib/mesos --master={ip}:5050 --attributes=preemptible:{preemptible} --no-hostname_lookup --no-systemd_enable_support"

        if self.clusterType == "mesos":
            if role == "leader":
                entryPoint = "mesos-master"
                entryPointArgs = MESOS_LOG_DIR + LEADER_DOCKER_ARGS.format(
                    name=self.clusterName
                )
            elif role == "worker":
                entryPoint = "mesos-agent"
                entryPointArgs = MESOS_LOG_DIR + WORKER_DOCKER_ARGS.format(
                    ip=self._leaderPrivateIP, preemptible=preemptible
                )
            else:
                raise RuntimeError("Unknown role %s" % role)
        elif self.clusterType == "kubernetes":
            if role == "leader":
                # We need *an* entry point or the leader container will finish
                # and go away, and thus not be available to take user logins.
                entryPoint = "sleep"
                entryPointArgs = "infinity"
            else:
                raise RuntimeError(
                    "Toil service not needed for %s nodes in a %s cluster",
                    role,
                    self.clusterType,
                )
        else:
            raise RuntimeError(
                "Toil service not needed in a %s cluster", self.clusterType
            )

        if keyPath:
            entryPointArgs = keyPath + " " + entryPointArgs
            entryPoint = "waitForKey.sh"
        customDockerInitCommand = customDockerInitCmd()
        if customDockerInitCommand:
            entryPointArgs = " ".join(
                ["'" + customDockerInitCommand + "'", entryPoint, entryPointArgs]
            )
            entryPoint = "customDockerInit.sh"

        # Set up the service. Make sure to make it default to using the
        # actually-big temp directory of /var/tmp (see
        # https://systemd.io/TEMPORARY_DIRECTORIES/).
        config.addUnit(
            f"toil-{role}.service",
            contents=textwrap.dedent(
                f"""\
            [Unit]
            Description=toil-{role} container
            After=docker.service
            After=create-kubernetes-cluster.service

            [Service]
            Restart=on-failure
            RestartSec=2
            ExecStartPre=-/usr/bin/docker rm toil_{role}
            ExecStartPre=-/usr/bin/bash -c '{customInitCmd()}'
            ExecStart=/usr/bin/docker run \\
                {self.toil_service_env_options()} \\
                --entrypoint={entryPoint} \\
                --net=host \\
                --init \\
                -v /var/run/docker.sock:/var/run/docker.sock \\
                -v /run/lock:/run/lock \\
                -v /var/lib/mesos:/var/lib/mesos \\
                -v /var/lib/docker:/var/lib/docker \\
                -v /var/lib/toil:/var/lib/toil \\
                -v /var/lib/cwl:/var/lib/cwl \\
                -v /var/tmp:/var/tmp \\
                -v /tmp:/tmp \\
                -v /opt:/opt \\
                -v /etc/kubernetes:/etc/kubernetes \\
                -v /etc/kubernetes/admin.conf:/root/.kube/config \\
                {"-e TOIL_KUBERNETES_PRIVILEGED=True --privileged" if self.enable_fuse else 
                "--security-opt seccomp=unconfined --security-opt systempaths=unconfined"} \\
                -e TOIL_KUBERNETES_HOST_PATH=/var/lib/toil \\
                # Pass in a path to use for singularity image caching into the container
                -e SINGULARITY_CACHEDIR=/var/lib/toil/singularity \\
                -e MINIWDL__SINGULARITY__IMAGE_CACHE=/var/lib/toil/miniwdl \\
                --name=toil_{role} \\
                {applianceSelf()} \\
                {entryPointArgs}

            [Install]
            WantedBy=multi-user.target
            """
            ),
        )

    def getKubernetesValues(self, architecture: str = "amd64"):
        """
        Returns a dict of Kubernetes component versions and paths for formatting into Kubernetes-related templates.
        """
        cloud_provider = self.getKubernetesCloudProvider()
        return dict(
            ARCHITECTURE=architecture,
            CNI_VERSION="v0.8.2",
            CRICTL_VERSION="v1.17.0",
            CNI_DIR="/opt/cni/bin",
            DOWNLOAD_DIR="/opt/bin",
            SETUP_STATE_DIR="/etc/toil/kubernetes",
            # This is the version of Kubernetes to use
            # Get current from: curl -sSL https://dl.k8s.io/release/stable.txt
            # Make sure it is compatible with the kubelet.service unit we ship, or update that too.
            KUBERNETES_VERSION="v1.19.3",
            # Now we need the basic cluster services
            # Version of Flannel networking to get the YAML from
            FLANNEL_VERSION="v0.13.0",
            # Version of node CSR signign bot to run
            RUBBER_STAMP_VERSION="v0.3.1",
            # Version of the autoscaler to run
            AUTOSCALER_VERSION="1.19.0",
            # Version of metrics service to install for `kubectl top nodes`
            METRICS_API_VERSION="v0.3.7",
            CLUSTER_NAME=self.clusterName,
            # YAML line that tells the Kubelet to use a cloud provider, if we need one.
            CLOUD_PROVIDER_SPEC=(
                ("cloud-provider: " + cloud_provider) if cloud_provider else ""
            ),
        )

    def addKubernetesServices(
        self, config: InstanceConfiguration, architecture: str = "amd64"
    ):
        """
        Add installing Kubernetes and Kubeadm and setting up the Kubelet to run when configured to an instance configuration.
        The same process applies to leaders and workers.
        """

        values = self.getKubernetesValues(architecture)

        # We're going to ship the Kubelet service from Kubernetes' release pipeline via cloud-config
        config.addUnit(
            "kubelet.service",
            contents=textwrap.dedent(
                """\
            # This came from https://raw.githubusercontent.com/kubernetes/release/v0.4.0/cmd/kubepkg/templates/latest/deb/kubelet/lib/systemd/system/kubelet.service
            # It has been modified to replace /usr/bin with {DOWNLOAD_DIR}
            # License: https://raw.githubusercontent.com/kubernetes/release/v0.4.0/LICENSE

            [Unit]
            Description=kubelet: The Kubernetes Node Agent
            Documentation=https://kubernetes.io/docs/home/
            Wants=network-online.target
            After=network-online.target

            [Service]
            ExecStart={DOWNLOAD_DIR}/kubelet
            Restart=always
            StartLimitInterval=0
            RestartSec=10

            [Install]
            WantedBy=multi-user.target
            """
            ).format(**values),
        )

        # It needs this config file
        config.addFile(
            "/etc/systemd/system/kubelet.service.d/10-kubeadm.conf",
            mode="0644",
            contents=textwrap.dedent(
                """\
            # This came from https://raw.githubusercontent.com/kubernetes/release/v0.4.0/cmd/kubepkg/templates/latest/deb/kubeadm/10-kubeadm.conf
            # It has been modified to replace /usr/bin with {DOWNLOAD_DIR}
            # License: https://raw.githubusercontent.com/kubernetes/release/v0.4.0/LICENSE

            # Note: This dropin only works with kubeadm and kubelet v1.11+
            [Service]
            Environment="KUBELET_KUBECONFIG_ARGS=--bootstrap-kubeconfig=/etc/kubernetes/bootstrap-kubelet.conf --kubeconfig=/etc/kubernetes/kubelet.conf"
            Environment="KUBELET_CONFIG_ARGS=--config=/var/lib/kubelet/config.yaml"
            # This is a file that "kubeadm init" and "kubeadm join" generates at runtime, populating the KUBELET_KUBEADM_ARGS variable dynamically
            EnvironmentFile=-/var/lib/kubelet/kubeadm-flags.env
            # This is a file that the user can use for overrides of the kubelet args as a last resort. Preferably, the user should use
            # the .NodeRegistration.KubeletExtraArgs object in the configuration files instead. KUBELET_EXTRA_ARGS should be sourced from this file.
            EnvironmentFile=-/etc/default/kubelet
            ExecStart=
            ExecStart={DOWNLOAD_DIR}/kubelet $KUBELET_KUBECONFIG_ARGS $KUBELET_CONFIG_ARGS $KUBELET_KUBEADM_ARGS $KUBELET_EXTRA_ARGS
            """
            ).format(**values),
        )

        # Before we let the kubelet try to start, we have to actually download it (and kubeadm)
        # We set up this service so it can restart on failure despite not
        # leaving a process running, see
        # <https://github.com/openshift/installer/pull/604> and
        # <https://github.com/litew/droid-config-ham/commit/26601d85d9d972dc1560096db1c419fd6fd9b238>
        # We use a forking service with RemainAfterExit, since that lets
        # restarts work if the script fails. We also use a condition which
        # treats the service as successful and skips it if it made a file to
        # say it already ran.
        config.addFile(
            "/home/core/install-kubernetes.sh",
            contents=textwrap.dedent(
                """\
            #!/usr/bin/env bash
            set -e
            FLAG_FILE="{SETUP_STATE_DIR}/install-kubernetes.done"

            # Make sure we have Docker enabled; Kubeadm later might complain it isn't.
            systemctl enable docker.service

            mkdir -p {CNI_DIR}
            curl -L "https://github.com/containernetworking/plugins/releases/download/{CNI_VERSION}/cni-plugins-linux-{ARCHITECTURE}-{CNI_VERSION}.tgz" | tar -C {CNI_DIR} -xz
            mkdir -p {DOWNLOAD_DIR}
            curl -L "https://github.com/kubernetes-sigs/cri-tools/releases/download/{CRICTL_VERSION}/crictl-{CRICTL_VERSION}-linux-{ARCHITECTURE}.tar.gz" | tar -C {DOWNLOAD_DIR} -xz

            cd {DOWNLOAD_DIR}
            curl -L --remote-name-all https://storage.googleapis.com/kubernetes-release/release/{KUBERNETES_VERSION}/bin/linux/{ARCHITECTURE}/{{kubeadm,kubelet,kubectl}}
            chmod +x {{kubeadm,kubelet,kubectl}}

            mkdir -p "{SETUP_STATE_DIR}"
            touch "$FLAG_FILE"
            """
            ).format(**values),
        )
        config.addUnit(
            "install-kubernetes.service",
            contents=textwrap.dedent(
                """\
            [Unit]
            Description=base Kubernetes installation
            Wants=network-online.target
            After=network-online.target
            Before=kubelet.service
            ConditionPathExists=!{SETUP_STATE_DIR}/install-kubernetes.done

            [Service]
            ExecStart=/usr/bin/bash /home/core/install-kubernetes.sh
            Type=forking
            RemainAfterExit=yes
            Restart=on-failure
            RestartSec=5s

            [Install]
            WantedBy=multi-user.target
            RequiredBy=kubelet.service
            """
            ).format(**values),
        )

        # Now we should have the kubeadm command, and the bootlooping kubelet
        # waiting for kubeadm to configure it.

    def getKubernetesAutoscalerSetupCommands(self, values: dict[str, str]) -> str:
        """
        Return Bash commands that set up the Kubernetes cluster autoscaler for
        provisioning from the environment supported by this provisioner.

        Should only be implemented if Kubernetes clusters are supported.

        :param values: Contains definitions of cluster variables, like
                       AUTOSCALER_VERSION and CLUSTER_NAME.

        :returns: Bash snippet
        """
        raise NotImplementedError()

    def getKubernetesCloudProvider(self) -> Optional[str]:
        """
        Return the Kubernetes cloud provider (for example, 'aws'), to pass to
        the kubelets in a Kubernetes cluster provisioned using this provisioner.

        Defaults to None if not overridden, in which case no cloud provider
        integration will be used.

        :returns: Cloud provider name, or None
        """
        return None

    def addKubernetesLeader(self, config: InstanceConfiguration):
        """
        Add services to configure as a Kubernetes leader, if Kubernetes is already set to be installed.
        """

        values = self.getKubernetesValues()

        # Customize scheduler to pack jobs into as few nodes as possible
        # See: https://kubernetes.io/docs/reference/scheduling/config/#profiles
        config.addFile(
            "/home/core/scheduler-config.yml",
            mode="0644",
            contents=textwrap.dedent(
                """\
            apiVersion: kubescheduler.config.k8s.io/v1beta1
            kind: KubeSchedulerConfiguration
            clientConnection:
              kubeconfig: /etc/kubernetes/scheduler.conf
            profiles:
              - schedulerName: default-scheduler
                plugins:
                  score:
                    disabled:
                    - name: NodeResourcesLeastAllocated
                    enabled:
                    - name: NodeResourcesMostAllocated
                      weight: 1
            """.format(
                    **values
                )
            ),
        )

        # Main kubeadm cluster configuration.
        # Make sure to mount the scheduler config where the scheduler can see
        # it, which is undocumented but inferred from
        # https://pkg.go.dev/k8s.io/kubernetes@v1.21.0/cmd/kubeadm/app/apis/kubeadm#ControlPlaneComponent
        config.addFile(
            "/home/core/kubernetes-leader.yml",
            mode="0644",
            contents=textwrap.dedent(
                """\
            apiVersion: kubeadm.k8s.io/v1beta2
            kind: InitConfiguration
            nodeRegistration:
              kubeletExtraArgs:
                volume-plugin-dir: "/opt/libexec/kubernetes/kubelet-plugins/volume/exec/"
                {CLOUD_PROVIDER_SPEC}
            ---
            apiVersion: kubeadm.k8s.io/v1beta2
            kind: ClusterConfiguration
            controllerManager:
              extraArgs:
                flex-volume-plugin-dir: "/opt/libexec/kubernetes/kubelet-plugins/volume/exec/"
            scheduler:
              extraArgs:
                config: "/etc/kubernetes/scheduler-config.yml"
              extraVolumes:
                - name: schedulerconfig
                  hostPath: "/home/core/scheduler-config.yml"
                  mountPath: "/etc/kubernetes/scheduler-config.yml"
                  readOnly: true
                  pathType: "File"
            networking:
              serviceSubnet: "10.96.0.0/12"
              podSubnet: "10.244.0.0/16"
              dnsDomain: "cluster.local"
            ---
            apiVersion: kubelet.config.k8s.io/v1beta1
            kind: KubeletConfiguration
            serverTLSBootstrap: true
            rotateCertificates: true
            cgroupDriver: systemd
            """.format(
                    **values
                )
            ),
        )

        # Make a script to apply that and the other cluster components
        # Note that we're escaping {{thing}} as {{{{thing}}}} because we need to match mustaches in a yaml we hack up.
        config.addFile(
            "/home/core/create-kubernetes-cluster.sh",
            contents=textwrap.dedent(
                """\
            #!/usr/bin/env bash
            set -e

            FLAG_FILE="{SETUP_STATE_DIR}/create-kubernetes-cluster.done"

            export PATH="$PATH:{DOWNLOAD_DIR}"

            # We need the kubelet being restarted constantly by systemd while kubeadm is setting up.
            # Systemd doesn't really let us say that in the unit file.
            systemctl start kubelet

            # We also need to set the hostname for 'kubeadm init' to work properly.
            /bin/sh -c "/usr/bin/hostnamectl set-hostname $(curl -s http://169.254.169.254/latest/meta-data/hostname)"

            if [[ ! -e /etc/kubernetes/admin.conf ]] ; then
                # Only run this once, it isn't idempotent
                kubeadm init --config /home/core/kubernetes-leader.yml
            fi

            mkdir -p $HOME/.kube
            cp /etc/kubernetes/admin.conf $HOME/.kube/config
            chown $(id -u):$(id -g) $HOME/.kube/config

            # Install network
            kubectl apply -f https://raw.githubusercontent.com/flannel-io/flannel/{FLANNEL_VERSION}/Documentation/kube-flannel.yml

            # Deploy rubber stamp CSR signing bot
            kubectl apply -f https://raw.githubusercontent.com/kontena/kubelet-rubber-stamp/release/{RUBBER_STAMP_VERSION}/deploy/service_account.yaml
            kubectl apply -f https://raw.githubusercontent.com/kontena/kubelet-rubber-stamp/release/{RUBBER_STAMP_VERSION}/deploy/role.yaml
            kubectl apply -f https://raw.githubusercontent.com/kontena/kubelet-rubber-stamp/release/{RUBBER_STAMP_VERSION}/deploy/role_binding.yaml
            kubectl apply -f https://raw.githubusercontent.com/kontena/kubelet-rubber-stamp/release/{RUBBER_STAMP_VERSION}/deploy/operator.yaml

            """
            ).format(**values)
            + self.getKubernetesAutoscalerSetupCommands(values)
            + textwrap.dedent(
                """\
            # Set up metrics server, which needs serverTLSBootstrap and rubber stamp, and insists on running on a worker
            curl -sSL https://github.com/kubernetes-sigs/metrics-server/releases/download/{METRICS_API_VERSION}/components.yaml | \\
                sed 's/          - --secure-port=4443/          - --secure-port=4443\\n          - --kubelet-preferred-address-types=Hostname/' | \\
                kubectl apply -f -

            # Grab some joining info and make a file we can parse later with configparser
            echo "[DEFAULT]" >/etc/kubernetes/worker.ini
            echo "JOIN_TOKEN=$(kubeadm token create --ttl 0)" >>/etc/kubernetes/worker.ini
            echo "JOIN_CERT_HASH=sha256:$(openssl x509 -pubkey -in /etc/kubernetes/pki/ca.crt | openssl rsa -pubin -outform der 2>/dev/null | openssl dgst -sha256 -hex | sed 's/^.* //')" >>/etc/kubernetes/worker.ini
            echo "JOIN_ENDPOINT=$(hostname):6443" >>/etc/kubernetes/worker.ini

            mkdir -p "{SETUP_STATE_DIR}"
            touch "$FLAG_FILE"
            """
            ).format(**values),
        )
        config.addUnit(
            "create-kubernetes-cluster.service",
            contents=textwrap.dedent(
                """\
            [Unit]
            Description=Kubernetes cluster bootstrap
            After=install-kubernetes.service
            After=docker.service
            Before=toil-leader.service
            # Can't be before kubelet.service because Kubelet has to come up as we run this.
            ConditionPathExists=!{SETUP_STATE_DIR}/create-kubernetes-cluster.done

            [Service]
            ExecStart=/usr/bin/bash /home/core/create-kubernetes-cluster.sh
            Type=forking
            RemainAfterExit=yes
            Restart=on-failure
            RestartSec=5s

            [Install]
            WantedBy=multi-user.target
            RequiredBy=toil-leader.service
            """
            ).format(**values),
        )

        # We also need a node cleaner service
        config.addFile(
            "/home/core/cleanup-nodes.sh",
            contents=textwrap.dedent(
                """\
            #!/usr/bin/env bash
            # cleanup-nodes.sh: constantly clean up NotReady nodes that are tainted as having been deleted
            set -e

            export PATH="$PATH:{DOWNLOAD_DIR}"

            while true ; do
                echo "$(date | tr -d '\\n'): Checking for scaled-in nodes..."
                for NODE_NAME in $(kubectl --kubeconfig /etc/kubernetes/admin.conf get nodes -o json | jq -r '.items[] | select(.spec.taints) | select(.spec.taints[] | select(.key == "ToBeDeletedByClusterAutoscaler")) | select(.spec.taints[] | select(.key == "node.kubernetes.io/unreachable")) | select(.status.conditions[] | select(.type == "Ready" and .status == "Unknown")) | .metadata.name' | tr '\\n' ' ') ; do
                    # For every node that's tainted as ToBeDeletedByClusterAutoscaler, and
                    # as node.kubernetes.io/unreachable, and hasn't dialed in recently (and
                    # is thus in readiness state Unknown)
                    echo "Node $NODE_NAME is supposed to be scaled away and also gone. Removing from cluster..."
                    # Drop it if possible
                    kubectl --kubeconfig /etc/kubernetes/admin.conf delete node "$NODE_NAME" || true
                done
                sleep 300
            done
            """
            ).format(**values),
        )
        config.addUnit(
            "cleanup-nodes.service",
            contents=textwrap.dedent(
                """\
            [Unit]
            Description=Remove scaled-in nodes
            After=create-kubernetes-cluster.service
            Requires=create-kubernetes-cluster.service
            [Service]
            ExecStart=/home/core/cleanup-nodes.sh
            Restart=always
            StartLimitInterval=0
            RestartSec=10
            [Install]
            WantedBy=multi-user.target
            """
            ),
        )

    def addKubernetesWorker(
        self,
        config: InstanceConfiguration,
        authVars: dict[str, str],
        preemptible: bool = False,
    ):
        """
        Add services to configure as a Kubernetes worker, if Kubernetes is
        already set to be installed.

        Authenticate back to the leader using the JOIN_TOKEN, JOIN_CERT_HASH,
        and JOIN_ENDPOINT set in the given authentication data dict.

        :param config: The configuration to add services to
        :param authVars: Dict with authentication info
        :param preemptible: Whether the worker should be labeled as preemptible or not
        """

        # Collect one combined set of auth and general settings.
        values = dict(**self.getKubernetesValues(), **authVars)

        # Mark the node as preemptible if it is.
        # TODO: We use the same label that EKS uses here, because nothing is standardized.
        # This won't be quite appropriate as we aren't on EKS and we might not
        # even be on AWS, but the batch system should understand it.
        values["WORKER_LABEL_SPEC"] = (
            'node-labels: "eks.amazonaws.com/capacityType=SPOT"' if preemptible else ""
        )

        # Kubeadm worker configuration
        config.addFile(
            "/home/core/kubernetes-worker.yml",
            mode="0644",
            contents=textwrap.dedent(
                """\
            apiVersion: kubeadm.k8s.io/v1beta2
            kind: JoinConfiguration
            nodeRegistration:
              kubeletExtraArgs:
                volume-plugin-dir: "/opt/libexec/kubernetes/kubelet-plugins/volume/exec/"
                {CLOUD_PROVIDER_SPEC}
                {WORKER_LABEL_SPEC}
            discovery:
              bootstrapToken:
                apiServerEndpoint: {JOIN_ENDPOINT}
                token: {JOIN_TOKEN}
                caCertHashes:
                - "{JOIN_CERT_HASH}"
            ---
            apiVersion: kubelet.config.k8s.io/v1beta1
            kind: KubeletConfiguration
            cgroupDriver: systemd
            """.format(
                    **values
                )
            ),
        )

        # Make a script to join the cluster using that configuration
        config.addFile(
            "/home/core/join-kubernetes-cluster.sh",
            contents=textwrap.dedent(
                """\
            #!/usr/bin/env bash
            set -e
            FLAG_FILE="{SETUP_STATE_DIR}/join-kubernetes-cluster.done"

            export PATH="$PATH:{DOWNLOAD_DIR}"

            # We need the kubelet being restarted constantly by systemd while kubeadm is setting up.
            # Systemd doesn't really let us say that in the unit file.
            systemctl start kubelet

            kubeadm join {JOIN_ENDPOINT} --config /home/core/kubernetes-worker.yml

            mkdir -p "{SETUP_STATE_DIR}"
            touch "$FLAG_FILE"
            """
            ).format(**values),
        )

        config.addUnit(
            "join-kubernetes-cluster.service",
            contents=textwrap.dedent(
                """\
            [Unit]
            Description=Kubernetes cluster membership
            After=install-kubernetes.service
            After=docker.service
            # Can't be before kubelet.service because Kubelet has to come up as we run this.
            Requires=install-kubernetes.service
            ConditionPathExists=!{SETUP_STATE_DIR}/join-kubernetes-cluster.done

            [Service]
            ExecStart=/usr/bin/bash /home/core/join-kubernetes-cluster.sh
            Type=forking
            RemainAfterExit=yes
            Restart=on-failure
            RestartSec=5s

            [Install]
            WantedBy=multi-user.target
            """
            ).format(**values),
        )

    def _getIgnitionUserData(
        self,
        role: str,
        keyPath: Optional[str] = None,
        preemptible: bool = False,
        architecture: str = "amd64",
    ) -> str:
        """
        Return the text (not bytes) user data to pass to a provisioned node.

        If leader-worker authentication is currently stored, uses it to connect
        the worker to the leader.

        :param str keyPath: The path of a secret key for the worker to wait for the leader to create on it.
        :param bool preemptible: Set to true for a worker node to identify itself as preemptible in the cluster.
        """

        # Start with a base config
        config = self.getBaseInstanceConfiguration()

        if self.clusterType == "kubernetes":
            # Install Kubernetes
            self.addKubernetesServices(config, architecture)

            if role == "leader":
                # Set up the cluster
                self.addKubernetesLeader(config)

            # We can't actually set up a Kubernetes worker without credentials
            # to connect back to the leader.

        if self.clusterType == "mesos" or role == "leader":
            # Leaders, and all nodes in a Mesos cluster, need a Toil service
            self.add_toil_service(config, role, keyPath, preemptible)

        if role == "worker" and self._leaderWorkerAuthentication is not None:
            # We need to connect the worker to the leader.
            if self.clusterType == "mesos":
                # This involves an SSH public key form the leader
                config.addSSHRSAKey(self._leaderWorkerAuthentication)
            elif self.clusterType == "kubernetes":
                # We can install the Kubernetes worker and make it phone home
                # to the leader.
                # TODO: this puts sufficient info to fake a malicious worker
                # into the worker config, which probably is accessible by
                # anyone in the cloud account.
                self.addKubernetesWorker(
                    config, self._leaderWorkerAuthentication, preemptible=preemptible
                )

        # Make it into a string for Ignition
        user_data = config.toIgnitionConfig()

        # Check if the config size exceeds the user data limit. If so, we'll
        # write it to the cloud and let Ignition fetch it during startup.

        user_data_limit: int = self._get_user_data_limit()

        if len(user_data) > user_data_limit:
            logger.warning(
                f"Ignition config size exceeds the user data limit ({len(user_data)} > {user_data_limit}).  "
                "Writing to cloud storage..."
            )

            src = self._write_file_to_cloud(
                f"configs/{role}/config-{uuid4()}.ign",
                contents=user_data.encode("utf-8"),
            )

            return json.dumps(
                {
                    "ignition": {
                        "version": "2.2.0",
                        # See: https://github.com/coreos/ignition/blob/spec2x/doc/configuration-v2_2.md
                        "config": {
                            "replace": {
                                "source": src,
                            }
                        },
                    }
                },
                separators=(",", ":"),
            )

        return user_data



import os

from toil.common import Toil
from toil.job import Job
from toil.lib.io import mkdtemp


class HelloWorld(Job):
    def __init__(self, id):
        Job.__init__(self)
        self.inputFileID = id

    def run(self, fileStore):
        with fileStore.readGlobalFileStream(self.inputFileID, encoding="utf-8") as fi:
            with fileStore.writeGlobalFileStream(encoding="utf-8") as (
                fo,
                outputFileID,
            ):
                fo.write(fi.read() + "World!")
        return outputFileID


if __name__ == "__main__":
    jobstore: str = mkdtemp("tutorial_staging")
    os.rmdir(jobstore)
    options = Job.Runner.getDefaultOptions(jobstore)
    options.logLevel = "INFO"
    options.clean = "always"

    with Toil(options) as toil:
        if not toil.options.restart:
            ioFileDirectory = os.path.join(
                os.path.dirname(os.path.abspath(__file__)), "stagingExampleFiles"
            )
            inputFileID = toil.importFile(
                "file://" + os.path.abspath(os.path.join(ioFileDirectory, "in.txt"))
            )
            outputFileID = toil.start(HelloWorld(inputFileID))
        else:
            outputFileID = toil.restart()

        toil.exportFile(
            outputFileID,
            "file://" + os.path.abspath(os.path.join(ioFileDirectory, "out.txt")),
        )



from typing import Any, List, Sequence, Tuple

from _typeshed import Incomplete

from .callables import UserListener
from .listener import IListenerExcHandler
from .notificationmgr import INotificationHandler
from .topicargspec import ArgSpecGiven
from .topicdefnprovider import ITopicDefnProvider
from .topicexc import TopicDefnError as TopicDefnError
from .topicexc import TopicNameError as TopicNameError
from .topicobj import Topic
from .topicutils import ALL_TOPICS as ALL_TOPICS

class TreeConfig:
    notificationMgr: Incomplete
    listenerExcHandler: Incomplete
    raiseOnTopicUnspecified: bool
    def __init__(
        self,
        notificationHandler: INotificationHandler = ...,
        listenerExcHandler: IListenerExcHandler = ...,
    ) -> None: ...

class TopicManager:
    TOPIC_SPEC_NOT_SPECIFIED: int
    TOPIC_SPEC_ALREADY_CREATED: int
    TOPIC_SPEC_ALREADY_DEFINED: int
    def __init__(self, treeConfig: TreeConfig = ...) -> None: ...
    def getRootAllTopics(self) -> Topic: ...
    def addDefnProvider(
        self, providerOrSource: Any, format: Incomplete | None = ...
    ) -> ITopicDefnProvider: ...
    def clearDefnProviders(self) -> None: ...
    def getNumDefnProviders(self) -> int: ...
    def getTopic(self, name: str, okIfNone: bool = ...) -> Topic: ...
    def getOrCreateTopic(
        self, name: str, protoListener: UserListener = ...
    ) -> Topic: ...
    def isTopicInUse(self, name: str) -> bool: ...
    def hasTopicDefinition(self, name: str) -> bool: ...
    def checkAllTopicsHaveMDS(self) -> None: ...
    def delTopic(self, name: str) -> bool: ...
    def getTopicsSubscribed(self, listener: UserListener) -> List[Topic]: ...
    def clearTree(self) -> None: ...

class _MasterTopicDefnProvider:
    def __init__(self, treeConfig: TreeConfig) -> None: ...
    def addProvider(self, provider: ITopicDefnProvider) -> None: ...
    def clear(self) -> None: ...
    def getNumProviders(self) -> int: ...
    def getDefn(self, topicNameTuple: Sequence[str]) -> Tuple[str, ArgSpecGiven]: ...
    def isDefined(self, topicNameTuple: Sequence[str]) -> bool: ...
